package EFS::CLI::DistStats;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Readonly;
use Params::Validate qw( :all );

Readonly my $SCREEN_WIDTH => 1024;

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        history => { type => SCALAR },
        detail  => { type => SCALAR, optional => 1 },
        sortby  => { type => SCALAR, default => q{what}, optional => 1 },
    });

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    if ( not defined $history->{endtime} ) {
        warn(
            q{The specified history record appears to },
            qq{correspond to a command\n},
            qq{which has not yet completed, or which may have crashed.\n}
        );
    }

    my @diststats = $history->diststats;

    if ( not @diststats ) {
        die qq{No diststats for history id = '$args{history}'\n};
    }

    my $sortby = $args{sortby};
    $sortby    = q{where} if $sortby eq q{cell};

    my @rows = ();

    foreach my $diststats ( @diststats ) {

        if ( not defined $diststats->{endtime} ) {
            # This one's in progress, or stale (crashed process)
            next;
        }

        my $key = $diststats->{metaproj};
        $key .= qq{/$diststats->{project}} if $diststats->{project};
        $key .= qq{/$diststats->{release}} if $diststats->{release};
        $key .= qq{/$diststats->{install}} if $diststats->{install};

        my ($dest_cell) = (
            $diststats->{dest_cell} =~ m/\A (\w+ \. \w+) \./msx
        );

        push( @rows, {
            what  => $key,
            where => $dest_cell,
            files => $diststats->{file_xfer_count},
            bytes => $diststats->{bytes_received} + $diststats->{bytes_sent},
            rate  => sprintf( q{%.2f}, $diststats->{bytes_rate} ),
        });

     }

    if ( not @rows ) {
        die(
            qq{None of the diststats entries cointain any data (yet).\n},
            qq{If the command is still running, try again later.\n}
        );
    }

    my $table = Text::FormatTable->new(q{l  l  r  r  r});
    $table->head(q{What},q{Where},q{Files},q{Bytes},q{Rate (b/s)});
    $table->rule(q{=});

    my @sorted;

    if ( $sortby eq q{what} || $sortby eq q{where} ) {
        @sorted = sort { lc($a->{$sortby}) cmp lc($b->{$sortby}) } @rows;
    } else {
        @sorted = sort { $a->{$sortby} <=> $b->{$sortby} } @rows;
    }

    foreach my $row ( @sorted ) {
        $table->row( $row->{what}, $row->{where},
                     $row->{files}, $row->{bytes}, $row->{rate} );
    }

    STDOUT->print( $table->render($SCREEN_WIDTH) );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
