package EFS::CLI::History;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Readonly;
use Date::Calc qw( Delta_DHMS );
use Date::Parse qw( str2time );
use Params::Validate qw( :all );
use List::MoreUtils qw( any uniq );
use Try::Tiny;

Readonly my $SCREEN_WIDTH => 1024;

# XXX: This is ugly....
my @other_keys     = @EFS::DBI::Result::History::other_keys;
my @history_values = @EFS::DBI::Result::History::history_values;

our %validation_specs = (

    # You can filter the history on any of these history values:
    id         => { type => SCALAR,  optional => 1 },
    parent     => { type => SCALAR,  optional => 1 },
    starttime  => { type => SCALAR,  optional => 1 },
    endtime    => { type => SCALAR,  optional => 1 },
    startdate  => { type => SCALAR,  optional => 1 },
    enddate    => { type => SCALAR,  optional => 1 },
    byuser     => { type => SCALAR,  optional => 1 },
    bygroup    => { type => SCALAR,  optional => 1 },
    efshost    => { type => SCALAR,  optional => 1 },
    efsversion => { type => SCALAR,  optional => 1 },
    request    => { type => SCALAR,  optional => 1 },
    command    => { type => SCALAR,  optional => 1 },
    status     => { type => BOOLEAN, optional => 1 },

    # These options control the output
    all     => { type => BOOLEAN, optional => 1 },
    count   => { type => SCALAR,  optional => 1 },
    orderby => { type => SCALAR,  default  => q{id}, optional => 1 },
    verbose => { type => BOOLEAN, optional => 1 },
    report  => { type => SCALAR,  default  => q{default}, optional => 1 },
    days    => { type => SCALAR,  optional => 1 },
    weeks   => { type => SCALAR,  optional => 1 },
    help    => { type => BOOLEAN, optional => 1 },
    failed  => { type => BOOLEAN, optional => 1 },
);

sub history {

    my $class = shift;

    my (%args) = @_;

    Readonly my $HISTORY_DEFAULT => 100;

    my %search = ();

    if ( $args{byuser} || $args{bygroup} ) {

        my @authusers;

        if ( $args{bygroup} ) {
            foreach my $bygroup ( split m{[,\s]+}msx, $args{bygroup} ) {
                my $bygroup_obj = EFS::Groupname->new({ name => $bygroup });
                if ( my @members = $bygroup_obj->members ) {
                    push @authusers, @members;
                }
                else {
                    STDOUT->print(qq{Authgroup '$bygroup' has no members\n});
                }
            }
        }

        if ( $args{byuser} ) {
            foreach my $byuser ( split m{[,\s]+}msx, $args{byuser} ) {
                my $byuser_obj = EFS::Username->new({ name => $byuser });
                push @authusers, $byuser;
            }
        }

        if ( not @authusers ) {
            die qq{Specification of authusers expanded to an empty list\n};
        }

        $search{authuser} = [ uniq @authusers ];

    }

    if ( $args{failed} ) {
        if ( exists $args{status} ) {
            die qq{Only one of -status or -failed may be specified\n};
        }
        $search{status} = [ 0, undef ];
    }

    foreach my $numeric (qw( id parent status )) {
        next unless exists $args{$numeric};
        $search{$numeric} = $args{$numeric};
    }

    foreach my $like (qw( request )) {
        next unless exists $args{$like};
        my $need_escape = $args{$like} =~ s/([%_])/\\$1/gmsx ? 1 : 0;
        $args{$like} =~ tr!*?!%_!;

        #         push(@where,qq{ $like like '$args{$like}' } .
        #                  ( $need_escape ? qq{ escape '\\'} : '' ));
        # XXX: How do we do this with SQL::Abstract syntax?
        # Note sure about the escape part at all.
        $search{$like} = { q{like}, $args{$like} };
    }

    foreach my $varchar ( @other_keys, @history_values ) {

        next if grep { $varchar eq $_ } qw( status request );  # Handled above

        next unless exists $args{$varchar};

        my @varchar = split(m/,/msx, $args{$varchar});

        if ( $#varchar > 0 ) {
            $search{$varchar} = \@varchar;
        }
        else {
            $search{$varchar} = $args{$varchar};
        }

    }

    foreach my $pair ( [qw(starttime startdate)], [qw(endtime enddate)],
        [qw(days weeks)], ) {
        if ( $args{ $pair->[0] } && $args{ $pair->[1] } ) {
            die qq{The $pair->[0] and $pair->[1] options are mutually exclusive\n};
        }
    }

    my %compare = ( starttime => q{>=}, endtime => q{<=}, );

    foreach my $datetime (qw( starttime endtime startdate enddate )) {

        next unless exists $args{$datetime};

        my $column = $datetime;
        my $type = q{time};

        if ( $datetime =~ m{date$}msx ) {
            $type = q{date};
            $column =~ s/date/time/msx;
        }

        my $time = str2time( $args{$datetime} ) ||
            die(qq{Unable to parse $datetime value '$args{$datetime}'\n});

        my $string = $class->time2dbstr(
            time  => $time,
            type  => $type,
        );

        $search{ $column } = {
            $compare{ $column } => $string,
        };

    }

    if ( defined $args{weeks} ) {
        Readonly my $DAYS_IN_WEEK => 7;
        $args{days} = $DAYS_IN_WEEK * $args{weeks};
    }

    if ( defined $args{days} ) {
        Readonly my $SECONDS_IN_DAY => 60 * 60 * 24;

        my $string = $class->time2dbstr(
            time => ( time - ( $SECONDS_IN_DAY * abs($args{days}) ) ),
            type => q{date},
        );

        $search{starttime} = {
            $compare{starttime} => $string,
        };
    }

    my @columns = ( q{id}, q{starttime}, q{endtime}, q{status}, q{authuser}, q{request} );

    push ( @columns, qw( efshost efsversion logfile ) ) if $args{verbose};

    if ( not any { defined $args{$_} } qw(count weeks starttime startdate
                                          days) ) {
        $args{count} = $HISTORY_DEFAULT;
    }

    my %attrs = (
        order_by => { -desc => $args{orderby} },
        columns  => \@columns,
    );

    if ( not $args{all} and $args{count} ) {
        $attrs{rows} = $args{count};
    }

    my $history_rs = $class->get_resultset;
    my @history = $history_rs->search( \%search, \%attrs );

    if ( not @history ) {
        STDOUT->print( qq{No history records match the specified query.\n} );
        return 1;
    }

    my %report_methods = (

        default => sub {

            my (%args)    = @_;
            my (@history) = @{ $args{history} };

            my $table = Text::FormatTable->new( q{r  l  l  l  l  l}
                    . ( $args{verbose} ? q{  l  l  l} : q{} ) );
            $table->head(
                q{History ID},
                q{StartTime},
                q{Duration},
                q{Status},
                $args{verbose} ? ( q{EFSHost}, q{EFSVersion} ) : (),
                q{Authuser},
                q{Request},
                $args{verbose} ? (q{Logfile}) : ()
            );

            $table->rule;

            foreach my $history ( reverse @history ) {

                my $duration = $class->datediff(
                    starttime => $history->starttime,
                    endtime   => $history->endtime,
                );

                my $id        = $history->id;
                my $starttime = $history->starttime || q{undef};
                my $request   = $history->request || q{undef};
                my $authuser  = $history->authuser || q{undef};

                my $status = $history->status;
                $status = q{undef} unless defined $status;

                my $efshost
                    = $args{verbose} ? ( $history->efshost || q{undef} ) : q{};
                my $efsversion
                    = $args{verbose}
                    ? ( $history->efsversion || q{undef} )
                    : q{};
                my $logfile
                    = $args{verbose} ? ( $history->logfile || q{undef} ) : q{};

                my $starttime_show
                    = join ( q{ }, $starttime->ymd(q{/}), $starttime->hms );

                $table->row(
                    $id,
                    $starttime_show,
                    $duration,
                    $status,
                    $args{verbose} ? ( $efshost, $efsversion ) : (),
                    $authuser,
                    $history->request,
                    $args{verbose} ? ($logfile) : ()
                );

            }

            return STDOUT->print($table->render($SCREEN_WIDTH));

        },

        userstats => sub {

            my (%args)    = @_;
            my (@history) = @{ $args{history} };

            my $table = Text::FormatTable->new(q{l  l});
            $table->head( q{Authuser}, q{Count} );
            $table->rule;

            my %rec = ();
            my $authuser;

            foreach my $history (@history) {
                $rec{ $history->authuser }++;
            }

            if ( lc ( $args{orderby} ) =~ m/\A (efs)? user \z/msx ) {
                foreach my $user ( sort keys %rec ) {
                    $table->row( $user, $rec{$user} );
                }
            }
            else {
                foreach my $user ( reverse sort { $rec{$a} <=> $rec{$b} } keys %rec ) {
                    $table->row( $user, $rec{$user} );
                }
            }

            return STDOUT->print($table->render($SCREEN_WIDTH));

        },

        activity => sub {

            # XXX: This code is broken -- it doesn't do what it was
            # supposed to, so it is NOT documented in efs_history.pod
            # yet

            my (%args)    = @_;
            my (@history) = @{ $args{history} };

            my $table = Text::FormatTable->new(q{l  l  l  l});
            $table->head( q{Authuser}, q{Status}, q{Count}, q{Request} );
            $table->rule;

            my %rec = ();

            foreach my $history ( reverse @history ) {

                my $id        = $history->id;
                my $starttime = $history->starttime || q{undef};
                my $request   = $history->request || q{undef};
                my $authuser  = $history->authuser || q{undef};

                my $status      = $history->status;
                if ( not defined $status ) {
                    $status     = q{undef};
                }

                if ( $status eq q{1} ) {

                    if ( !defined $rec{$authuser}{good}{counter} ) {
                        $rec{$authuser}{good}{counter} = 1;
                        if ( defined $rec{$authuser}{failed}{$request} ) {
                            delete $rec{$authuser}{failed}{$request};
                            $rec{$authuser}{failed}{counter}--;
                        }
                    }
                    else {
                        $rec{$authuser}{good}{counter}++;
                        if ( defined $rec{$authuser}{failed}{$request} ) {
                            delete $rec{$authuser}{failed}{$request};
                            $rec{$authuser}{failed}{counter}--;
                        }
                    }
                }
                else {    # remember Failed request
                    if ( !defined $rec{$authuser}{failed}{counter} ) {

                        # the request $request appears first time
                        $rec{$authuser}{failed}{counter} = 1;
                    }
                    else {
                        if ( !defined $rec{$authuser}{failed}{$request} ) {

                            # increase total counter only if it is a
                            # new kind request
                            $rec{$authuser}{failed}{counter}++;
                        }
                    }
                    $rec{$authuser}{failed}{$request} = 1;
                }
            }

            foreach my $user ( keys %rec ) {
                if ( $rec{$user}{failed}{counter} ) {

                    if ( !$args{failed} ) {
                        $table->row( $user, q{Failed},
                                     $rec{$user}{failed}{counter},q{ });
                    }
                    else {

                        my $line = 0;
                        foreach my $req ( keys %{ $rec{$user}{failed} } ) {
                            next if ( $req eq q{counter} );
                            if ( $line++ == 0 ) {
                                $table->row( $user, q{Failed},
                                             $rec{$user}{failed}{counter},
                                             $req);
                            } else {
                                $table->row(q{ }, q{ }, q{ }, $req);
                            }
                        }
                    }
                }
                if ( $rec{$user}{good}{counter} ) {
                    $table->row( $user, q{Success},
                                 $rec{$user}{good}{counter},q{ } );
                }
            }

            return STDOUT->print($table->render($SCREEN_WIDTH));

        },

    );

    {
        ## no critic: ProhibitNoStrict
        no strict q{refs};
        $report_methods{ $args{report} }->( %args, history => \@history, );
    }

    return;

}

sub datediff {

    my $class = shift;

    my (%args) = validate( @_, {
        starttime => { type => OBJECT | UNDEF },
        endtime   => { type => OBJECT | UNDEF },
    });

    if ( not $args{starttime} or not $args{endtime} ) {
        return q{Unknown};
    }

    foreach my $field (qw( starttime endtime )) {
        return q{Unknown} if $args{$field}->year eq q{0000};
    }

    my $starttime = $args{starttime};
    my $endtime   = $args{endtime};

    my $starttime_show = join(q{ },$starttime->ymd,$starttime->hms);
    my $endtime_show   = join(q{ },$endtime->ymd,$endtime->hms);

    my ( $dday, $dhour, $dmin, $dsec );

    try {
        ( $dday, $dhour, $dmin, $dsec ) = Delta_DHMS(
            $starttime->year, $starttime->month,  $starttime->day,
            $starttime->hour, $starttime->minute, $starttime->second,
            $endtime->year,   $endtime->month,    $endtime->day,
            $endtime->hour,   $endtime->minute,   $endtime->second
        );
    }
    catch {
        die(
            qq{Unable to parse and compare start and end times:\n},
            qq{\tstarttime => '$args{starttime}'\n},
            qq{\tendtime   => '$args{endtime}'\n},
            $_,
        );
    };

    my @duration = ();

    # Note: we are trusting our code to never put out of order
    # start/endtimes into the database, otherwise, these values
    # will be negative.
    if ( $dday || $dhour || $dmin || $dsec ) {
        push(@duration,qq{${dday}d})  if $dday;
        push(@duration,qq{${dhour}h}) if $dhour;
        push(@duration,qq{${dmin}m})  if $dmin;
        push(@duration,qq{${dsec}s})  if $dsec;
    } elsif ( $args{endtime} ) {
        push(@duration,q{<1S});
    } else {
        push(@duration,q{Unknown});
    }

    return join(q{,},@duration);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
