package EFS::CLI::Home;

use strict;
use warnings;
use English;

use feature q{switch};

use base q{EFS::CLI};
use base q{EFS::DBI};

use EFS::FQDN qw( fqdn );

use Readonly;
use Params::Validate qw( :all );
use File::Basename qw( dirname );
use Date::Calc qw(:all);
use Date::Manip;
use Try::Tiny;

Readonly my $NUM_DAYS => 14;

sub create {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
        cell => { type => SCALAR },
    });

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my $cell_rs = $class->get_resultset(q{Cell});
    my $cell    = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    $cell->assert_cell_is_dev;
    $cell->assert_supports_filetype( q{home} );

    return EFS::RPC->execute(
        cell   => $cell,
        class  => $class,
        method => q{create_rpc},
        args   => [%args],
    );

}

sub create_rpc {

    my $class = shift;
    my %args  = @_;

    my $username = EFS::Username->new({ name => $args{name} });

    my $resultset = $class->get_resultset;
    my ($home)    = $resultset->search({ name => $args{name} });

    my $cell_rs = $class->get_resultset(q{Cell});
    my $cell    = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    if ($home) {

        if ( $home->destroyed ) {
            die(
                qq{The home '$home' has been destroyed.\n},
                qq{It must be recovered or purged.\n}
            );
        }

        if ( $home->cell->name ne $cell->name ) {
            die(qq{Home $home already exists in cell } . $home->cell . qq{\n});
        }

    }
    else {

        $home = $resultset->create({
            name => $args{name},
            cell => $cell->id,
        });

    }

    given ( $cell->fstype ) {

        when ( q{nfs} ) {

            my $admin_path = $cell->admin_path( type => q{home} );
            my $home_admin = $home->get_path_admin;

            EFS::Util->MakeDirHierarchy(
                top    => qq{$admin_path/home},
                bottom => dirname($home_admin),
            );

            EFS::Util->MakeDir(
                user  => $username->name,
                group => $username->primary_groupname,
                mode  => 02755,
                path  => $home_admin,
            );

            EFS::CLI::Namespace->addentry(
                type     => q{home},
                realpath => $home->object_id,
                autopath => $home->autopath,
            );

        }

        when ( q{afs} ) {
            die qq{Unable to create homes in afs cells yet\n};
        }

    }

    STDOUT->print( qq{Created home $home\n} );

    return $home;

}

sub move {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
        cell => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    return EFS::RPC->execute(
        cell   => $home->cell,
        class  => $class,
        method => q{move_rpc},
        args   => [%args],
    );

}

sub move_rpc {

    my $class = shift;

    my %args = @_;

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $resultset = $class->get_resultset;

    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    my $fromcell = $home->cell;

    my $cell_rs = $class->get_resultset(q{Cell});

    my $tocell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    if ( $fromcell->name eq $tocell->name ) {
        die(
            qq{Invalid argument: the target cell $tocell matches },
            qq{the home cell $fromcell\n},
        );
    }

    # XXX: Not sure how to handle AFS yet
    $tocell->assert_fstype( q{nfs} );

    $tocell->assert_cell_is_dev;
    $tocell->assert_supports_filetype( q{home} );

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my $home_lock = EFS::DirLock->new({
        directory => $home->get_path,
        retry     => 1,
        command   => $history->request,
    });

    # (1) remove /efs/home/a/ab/abcd from the namespace

    EFS::CLI::Namespace->delentry(
        type     => q{home},
        realpath => $home->object_id,
        showpath => $home->get_path,
    );

    # This entire step is try'd so we can call the move_backout
    # method is ANY of these steps fail.

    my $src_path      = $home->get_path_admin( $fromcell );

    my $dest_admin_path = $tocell->admin_path( type => q{home} );
    my $dest_path = $home->get_path_admin( $tocell );

    try {

        # (2) move the contents.

        EFS::RPC->execute(
            cell   => $tocell,
            class  => q{EFS::Util},
            method => q{MakeDirHierarchy},
            args   => [
                top    => $dest_admin_path,
                bottom => $dest_path,
            ],
        );

        # (2a) rsync from old to new cell

        my %rsync_args = (
            dest_cell => $tocell,
            src_path  => $src_path,
            dest_path => $dest_path,
        );

        my $first_rsync = EFS::Rsync->new( \%rsync_args );

        # (2b) rsync again, to verify that NOTHING was changed

        my $second_rsync = EFS::Rsync->new( \%rsync_args );

        my @updated = $second_rsync->get_updated;
        my @deleted = $second_rsync->get_deleted;

        if ( @updated || @deleted ) {

            my @message = (
                qq{Unable to move home $home, because the source contents\n},
                qq{have been modified.\n},
            );

            if (@updated) {
                push @message, (
                    qq{The following paths were updated:\n\t},
                    join ( qq{\n\t}, @updated ), qq{\n},
                );
            }

            if (@deleted) {
                push @message, (
                    qq{The following paths were deleted:\n\t},
                    join ( qq{\n\t}, @updated ), qq{\n},
                );
            }

            die(@message);

        }

    }
    catch {
        $home->move_backout(
            tocell  => $tocell,
            message => $_,
        );
    };

    # (3) Update the cell in the database.

    $home->set_column( cell => $tocell->id );
    $home->update;

    # (4) put /efs/home/a/ab/abcd back into the namespace

    EFS::CLI::Namespace->addentry(
        type     => q{home},
        realpath => $home->object_id,
        autopath => $home->autopath,
    );

    # (5) remove the source contents.

    EFS::Util->RemoveDirHierarchy( top => $src_path );

    STDOUT->print( qq{Successfully moved home $home to $tocell\n} );

    return 1;

}

sub moveall {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        fromcell => { type => SCALAR },
        tocell   => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset(q{Cell});

    my $fromcell = $cell_rs->retrieve_bynames_assert( name => $args{fromcell} );

    my $tocell = $cell_rs->retrieve_bynames_assert( name => $args{tocell} );

    $tocell->assert_cell_is_dev;

    # XXX: Not sure how to handle AFS yet
    $tocell->assert_fstype( q{nfs} );

    my @homes = grep { not $_->destroyed } $fromcell->homes;

    if ( not @homes ) {
        STDOUT->print( qq{No homes to move in cell $fromcell\n} );
        return 1;
    }

    my $fm_homes = EFS::ForkManager->new({

        list => \@homes,

        action => sub {

            my $home = shift;

            my $object_id = $home->object_id(q{ });

            my $tocell_name = $tocell->name;

            return EFS::RPC->execute(
                cell   => $fromcell,
                class  => $class,
                method => q{move_rpc},
                args   => [
                    name     => $home->name,
                    cell     => $tocell_name,
                ],
                history => {
                    parentid => $args{history},
                    values => {
                        request => qq{move home $object_id $tocell_name},
                        home => $home->name,
                    },
                },
            );

        },

    });

    $fm_homes->run || die(
        qq{Unable to move the following homes:\n\t},
        join ( qq{\n\t}, $fm_homes->get_failed ), qq{\n},
    );

    return 1;

}

sub destroy {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR },
    });

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my $resultset = $class->get_resultset;
    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    return EFS::RPC->execute(
        cell   => $home->cell,
        class  => $class,
        method => q{destroy_rpc},
        args   => [%args],
    );

}

sub destroy_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;
    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $home_dir = $home->get_path;

    my $home_lock;

    if ( -d $home_dir ) {
        $home_lock = EFS::DirLock->new({
            directory => $home_dir,
            retry     => 1,
            command   => $history->request,
        });
    }

    my $home_admin     = $home->get_path_admin;
    my $home_destroyed = $home->get_path_destroyed;

    

    EFS::Util->MakeDir(
        path => dirname($home_destroyed),
        mode => 0700,
    );

    EFS::Util->RenameDir(
        from  => $home_admin,
        to    => $home_destroyed,
    );

    EFS::CLI::Namespace->delentry(
        type     => q{home},
        realpath => $home->object_id,
        showpath => $home->get_path,
    );

    $home->set_column( destroyed => join(q{-}, Today) );
    $home->update;

    STDOUT->print(qq{Destroyed home $home\n});

    return 1;

}

sub destroy_cell {

    my $class = shift;
    my $home = shift;

    given ( $home->cell->fstype ) {

        when ( q{nfs} ) {

            my $home_admin = $home->get_path_admin;
            my $destroyed_dir = $home->get_path_destroyed;

            EFS::Util->MakeDir(
                path => dirname($destroyed_dir),
                mode => 0700,
            );

            EFS::Util->RenameDir(
                from  => $home_admin,
                to    => $destroyed_dir,
            );

        }

        when ( q{afs} ) {

            # XXX: This will involve just unmounting the home volume

            die qq{Unable to destroy homes in AFS cells yet\n};

        }

    }

    return 1;

}

sub purge {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR, optional => 1 },
        now  => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;
    my $cell_rs = $class->get_resultset( q{Cell} );

    if ( $args{name} ) {
        my ($cell) = $cell_rs->search({ name => $args{name} });
        if ( $cell ) {
            $args{cell} = delete $args{name};
        }
    }

    my @homes = ();

    if ( $args{name} ) {

        my $home = $homes[0] =
            $resultset->retrieve_bynames_assert( name => $args{name} );

        my $destroyed = $home->destroyed ||
            die(qq{Home $home has not been destroyed\n});

        if ( not $args{now} ) {

            if ( $home->delta_days( $destroyed ) < $NUM_DAYS ) {
                STDOUT->print(
                    qq{Unable to purge $home\n},
                    q{It was destroyed less than },
                    qq{$NUM_DAYS days ago ($destroyed).\n},
                );
                return 1;
            }

        }

    }
    else {

        if ( $args{cell} ) {
            my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
            @homes = grep { $_->destroyed } $resultset->search({ cell => $cell->id });
        }
        else {
            @homes = grep { $_->destroyed } $resultset->all;
        }

        if ( not @homes ) {
            STDOUT->print( qq{No homes to purge\n} );
            return 1;
        }

        my @purge = ();

        foreach my $home ( @homes ) {
            if ( $home->delta_days( $home->destroyed ) >= $NUM_DAYS ) {
                push @purge, $home;
            }
        }

        if ( not @purge ) {
            STDOUT->print( qq{No expired homes to purge\n} );
            return 1;
        }

        @homes = @purge;

    }

    my $home_fm = EFS::ForkManager->new({

        list => \@homes,

        action => sub {

            my $home = shift;

            return EFS::RPC->execute(
                cell   => $home->cell,
                class  => $class,
                method => q{purge_rpc},
                args   => [ $home->name ],
            );

        },

    });

    $home_fm->run || die(
        qq{Unable to purge the following homes:\n\t},
        join(qq{\n\t}, map { $_->object_id } $home_fm->get_failed ), qq{\n}
    );

    return 1;

}

sub purge_rpc {

    my $class = shift;
    my $name = shift;

    my $resultset = $class->get_resultset;
    my $home = $resultset->retrieve_bynames_assert( name => $name );

    given ( $home->cell->fstype ) {

        when ( q{nfs} ) {
            EFS::Util->RemoveDirHierarchy( top => $home->get_path_destroyed );
        }

        when ( q{afs} ) {
            # XXX: This will involve removing the actual home volume
            die qq{Unable to purge homes form AFS cells yet\n};
        }

    }

    $home->attrs->delete_all;
    $home->delete;

    STDOUT->print(qq{Successfully purged home $name\n});

    return 1;

}

sub recover {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;
    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    return EFS::RPC->execute(
        cell   => $home->cell,
        class  => $class,
        method => q{recover_rpc},
        args   => [%args],
    );

}

sub recover_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;
    my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    $home->destroyed || die qq{Home $home has not been destroyed\n};

    given ( $home->cell->fstype ) {

        when ( q{nfs} ) {

            my $home_admin = $home->get_path_admin;
            my $destroyed_dir = $home->get_path_destroyed;

            EFS::Util->RenameDir(
                from => $destroyed_dir,
                to   => $home_admin,
            );

        }

        when ( q{afs} ) {
            
            # XXX: This will involve remounting the home volume
            die qq{Unable to recover homes in AFS cells yet\n};

        }

    }

    EFS::CLI::Namespace->addentry(
        type     => q{home},
        realpath => $home->object_id,
        autopath => $home->autopath,
    );

    $home->set_column( destroyed => undef );
    $home->update;

    STDOUT->print( qq{Successfully recovered home $home\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;
    my $cell_rs = $class->get_resultset( q{Cell} );

    if ( $args{name} ) {
        my ($cell) = $cell_rs->search({ name => $args{name} });
        if ( $cell ) {
            $args{cell} = delete $args{name};
        }
    }

    if ( $args{name} ) {

        my $home = $resultset->retrieve_bynames_assert( name => $args{name} );

        if ( my $destroyed = $home->destroyed ) {
            STDOUT->print(
                qq{NOTICE: Home '$home' was destroyed on '$destroyed'\n}
            );
        }

        $class->display_objects( objects => [ $home ] );

        if ( my @attrs = $home->attrs ) {
            $attrs[0]->display_objects( objects => \@attrs );
        }

        return 1;

    }

    my @homes = ();

    if ( $args{cell} ) {
        my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        @homes = $resultset->search({ cell => $cell->id });
    }
    else {
        @homes = $resultset->all;
    }

    if ( not @homes ) {
        STDOUT->print(qq{No objects match the specified criteria\n});
        return 1;
    }

    my @active_homes = sort grep { not $_->destroyed } @homes;
    my @destroyed_homes = sort grep { $_->destroyed } @homes;

    if ( @active_homes ) {
        STDOUT->print( qq{Active Homes:\n} );
        $class->display_objects( objects => \@homes );
    }

    if ( @destroyed_homes ) {
        STDOUT->print( qq{Destroyed Homes\n} );
        $class->display_objects(
            objects => \@homes,
            display_columns => [qw(name cell destroyed)],
        );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        name     => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{home} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
