package EFS::CLI::Project;

use strict;
use warnings;
use Try::Tiny;
use English;

use feature q{switch};

use base q{EFS::CLI};
use base q{EFS::DBI};
use base q{EFS::DistableObject};

use EFS::FQDN qw( fqdn );

use Params::Validate qw( :all );
use File::Basename qw( dirname );

sub set_movelock {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        name      => { type => SCALAR },
        efsserver => { type => SCALAR },
        pid       => { type => SCALAR },
        fromcell  => { type => SCALAR },
        tocell    => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $efsserver_rs = $class->get_resultset(q{EFSServer});

    my $efsserver =
        $efsserver_rs->retrieve_bynames_assert( name => $args{efsserver} );

    my $cell_rs = $class->get_resultset(q{Cell});

    my $fromcell =
        $cell_rs->retrieve_bynames_assert( name => $args{fromcell} );

    my $tocell = $cell_rs->retrieve_bynames_assert( name => $args{tocell} );

    if ( $args{pid} !~ m/\A \d+ \z/msx ) {
        die qq{Invalid pid '$args{pid}'.  Must be numeric\n};
    }

    my $movelock = qq{$efsserver:$args{pid}:$fromcell:$tocell};

    $project->set_column( movelock => $movelock );
    $project->update;

    STDOUT->print( qq{Set movelock for $project to $movelock\n} );

    return 1;

}

sub clear_movelock {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
        movelock => 1,
    );

    $project->set_column( movelock => undef );
    $project->update;

    STDOUT->print( qq{Cleared movelock for $project\n} );

    return 1;
}

sub create {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dist_args_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        cell     => { type => SCALAR, optional => 1 },
        nocreate => { type => SCALAR, optional => 1 },
    });

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my @cells = $class->handle_cell_arguments(
        %args,
        celltype      => q{dev},
        default_scope => q{global},
    );

    my $metaproj_rs = $class->get_resultset(q{Metaproj});

    my $metaproj =
        $metaproj_rs->retrieve_bynames_assert( name => $args{metaproj} );

    my $cell = $class->get_localcell;

    if ( $args{cell} ) {
        my $cell_rs = $class->get_resultset(q{Cell});
        $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
    }

    if ( $cell->celltype ne q{dev} ) {
        die qq{Unable to create projects in non-dev cell '$cell'\n};
    }

    my $resultset = $class->get_resultset;

    my ($project) = $resultset->search({
        metaproj => $metaproj->id,
        name     => $args{name}
    });

    if ($project) {

        # This checks for disabled or movelocked projects
        $project = $resultset->retrieve_bynames_assert(
            metaproj => $metaproj->name,
            name     => $args{name},
        );

        if ( $project->cell->name ne $cell->name ) {
            die(
                qq{Project '$project' already exists },
                q{in cell '}, $project->cell, qq{'\n},
            );
        }

    }
    else {

        $project = $resultset->create({
            metaproj => $metaproj->id,
            name     => $args{name},
            cell     => $cell->id,
        });

    }

    if ( $args{nocreate} ) {
        STDOUT->print( qq{Defined project '$project' in cell '$cell'\n"} );
        return $project;
    }

    given ( $project->cell->fstype ) {

        when ( q{afs} ) {

            my $metaproj_devvolume = $metaproj->get_afsvolume( q{dev} );

            my $project_cell = $project->cell;

            my @partitions = $project_cell->get_partition_list;

            EFS::RPC->execute(
                cell   => $project_cell,
                class  => q{EFS::AFS::Util},
                method => q{create_volume},
                args   => [
                    name       => $metaproj_devvolume->name,
                    cell       => $project_cell->name,
                    partitions => \@partitions,
                    readonly   => 1,
                ],
            );

            EFS::RPC->execute(
                cell   => $project_cell,
                class  => q{EFS::AFS::Util},
                method => q{create_mount},
                args   => [
                    volume => $metaproj_devvolume->name,
                    parent => q{efs.dev},
                    path   => $metaproj->name,
                    cell   => $project_cell->name,
                ],
            );

            my $project_devvolume = $project->get_afsvolume( q{dev} );

            EFS::RPC->execute(
                cell   => $project_cell,
                class  => q{EFS::AFS::Util},
                method => q{create_volume},
                args   => [
                    name       => $project_devvolume->name,
                    cell       => $project_cell->name,
                    partitions => \@partitions,
                    readonly   => 1,
                ],
            );

            EFS::RPC->execute(
                cell   => $project_cell,
                class  => q{EFS::AFS::Util},
                method => q{create_mount},
                args   => [
                    volume => $project_devvolume->name,
                    parent => $metaproj_devvolume->name,
                    path   => $project->name,
                    cell   => $project_cell->name,
                ],
            );

        }

        when ( q{nfs} ) {

            my $admin_path = $project->cell->admin_path( type => q{dev} );

            EFS::RPC->execute(
                cell   => $project->cell,
                class  => q{EFS::Util},
                method => q{MakeDirHierarchy},
                args   => [
                    top    => qq{$admin_path/dev},
                    bottom => qq{$admin_path/dev/$project},
                ],
            );

        }
 
    }

    EFS::CLI::Namespace->addentry(
        type     => q{dev},
        cells    => \@cells,
        realpath => $project->object_id,
        autopath => $project->autopath,
    );

    STDOUT->print( qq{Created project $project\n} );

    my $project_dir = $project->get_path(q{dev});
    STDOUT->print( qq{_EFS_WAITFOR: $project_dir\n} );

    return $project;

}

sub move {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dist_args_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        cell     => { type => SCALAR },
        nobuild  => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{move_rpc},
        args   => [%args],
    );

}

sub move_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my $project_lock = EFS::DirLock->new({
        directory => $project->get_path( q{dev} ),
        retry     => 1,
        command   => $history->request,
    });

    my $fromcell = $project->cell;

    my $cell_rs = $class->get_resultset(q{Cell});
    my $tocell = $cell_rs->retrieve_bynames_assert( name => $args{cell}, );

    if ( $fromcell->fstype eq q{afs} or $tocell->fstype eq q{afs} ) {
        # XXX: This is going to take some work
        die qq{Unable to support move project to/from AFS cells yet\n};
    }

    if ( $fromcell->name eq $tocell->name ) {
        die(
            qq{Invalid argument: the target cell $tocell matches },
            qq{the project cell $fromcell\n},
        );
    }

    if ( $tocell->celltype ne q{dev} ) {
        die qq{Invalid argument: the target cell $tocell is not a dev cell\n};
    }

    # (1) remove /efs/dev/metaproj/project from the namespace

    my $movelock = join( q{:}, fqdn, $PID, $fromcell->name, $tocell->name );
    $project->set_column( movelock => $movelock );
    $project->update;

    EFS::CLI::Namespace->delentry(
        type     => q{dev},
        realpath => $project->object_id,
        showpath => $project->get_path(q{dev}),
    );

    # This entire step is try'd so we can call the move_backout
    # method is ANY of these steps fail.

    try {

        # (2) move the contents.

        my $src_admin_path = $fromcell->admin_path( type => q{dev} );
        my $src_path       = join( q{/},
                                   $src_admin_path,
                                   q{dev},
                                   $project->object_id );

        my $dest_admin_path = $tocell->admin_path( type => q{dev} );
        my $dest_path       = join( q{/},
                                    $dest_admin_path,
                                    q{dev},
                                    $project->object_id );

        my $dest_dirname = dirname($dest_path);

        EFS::RPC->execute(
            cell   => $tocell,
            class  => q{EFS::Util},
            method => q{MakeDirHierarchy},
            args   => [ top => $dest_dirname, bottom => $dest_path, ],
        );

        # (2a) rsync from old to new cell

        my %rsync_args = (
            dest_cell => $tocell,
            src_path  => $src_path,
            dest_path => $dest_path,
        );

        if ( $args{nobuild} ) {
            $rsync_args{exclude} = q{"/*/build/"};
        }

        my $first_rsync = EFS::Rsync->new( \%rsync_args );

        # (2b) rsync again, to verify that NOTHING was changed

        my $second_rsync = EFS::Rsync->new( \%rsync_args );

        my @updated = $second_rsync->get_updated;
        my @deleted = $second_rsync->get_deleted;

        if ( @updated || @deleted ) {

            my @message = (
                qq{Unable to move project $project, },
                qq{because the source contents\n},
                qq{have been modified.\n},
            );

            if (@updated) {
                push @message, (
                    qq{The following paths were updated:\n\t},
                    join ( qq{\n\t}, @updated ), qq{\n},
                );
            }

            if (@deleted) {
                push @message, (
                    qq{The following paths were deleted:\n\t},
                    join ( qq{\n\t}, @updated ), qq{\n},
                );
            }

            die @message;

        }

        # (2c) move the $platform/m/p directories, used to eval exec in
        # the localcell.  For these, we do NOT do two rsyncs.  They have
        # no real data, just symlinks.

        # NOTE: Parallelizing this with EFS::ForkManager exposes some
        # race conditions in the SSL authentication.

        foreach my $src_path ( glob( qq{$src_admin_path/*/dev/$project} ) ) {

            my $dest_path = $src_path;
            $dest_path    =~ s/\Q$src_admin_path\E/$dest_admin_path/msx;

            EFS::RPC->execute(
                cell   => $tocell,
                class  => q{EFS::Util},
                method => q{MakeDirHierarchy},
                args   => [
                    top    => $dest_admin_path,
                    bottom => $dest_path,
                ],
            );

            EFS::Rsync->new({
                dest_cell => $tocell,
                src_path  => $src_path,
                dest_path => $dest_path,
            });

        }

    }
    catch {
        $project->move_backout(
            tocell  => $tocell,
            message => $_,
        );
    };

    # (3) Update the cell in the database.

    $project->set_column( cell => $tocell->id );
    $project->update;

    # (4) put /efs/dev/metaproj/project back into the namespace

    $project->set_column( movelock => undef );
    $project->update;

    EFS::CLI::Namespace->addentry(
        type     => q{dev},
        realpath => $project->object_id,
        autopath => $project->autopath,
    );

    # (5) remove the source contents.  Note that this gets rid of the
    # project directory, and all of the "dotdirs", too.

    $class->destroy_cell(
        cell     => $fromcell->name,
        metaproj => $project->metaproj->name,
        name     => $project->name,
        type     => q{dev},
    );

    STDOUT->print( qq{Successfully moved project $project to $tocell\n} );

    return 1;

}

sub moveall {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        fromcell => { type => SCALAR },
        tocell   => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset(q{Cell});

    my $fromcell =
        $cell_rs->retrieve_bynames_assert( name => $args{fromcell} );

    my $tocell = $cell_rs->retrieve_bynames_assert( name => $args{tocell} );

    if ( $fromcell->fstype eq q{afs} or $tocell->fstype eq q{afs} ) {
        # XXX: This is going to take some work
        die qq{Unable to support move project to/from AFS cells yet\n};
    }

    if ( $tocell->celltype ne q{dev} ) {
        die qq{Invalid argument: tocell must be a dev cell\n};
    }

    my @projects = $class->sort_objects( $fromcell->projects );

    if ( not @projects ) {
        STDOUT->print( qq{No projects to move in cell $fromcell\n} );
        return 1;
    }

    my $fm_projects = EFS::ForkManager->new({

        list => \@projects,

        action => sub {

            my $project = shift;

            my $object_id = $project->object_id(q{ });

            my $tocell_name = $tocell->name;

            return EFS::RPC->execute(
                cell   => $fromcell,
                class  => $class,
                method => q{move_rpc},
                args   => [
                    metaproj => $project->metaproj->name,
                    name     => $project->name,
                    cell     => $tocell_name,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request => qq{move project $object_id $tocell_name},
                        project => $project->name,
                    },
                },
            );

        },

    });

    $fm_projects->run || die(
        qq{Unable to move the following projects:\n\t},
        join ( qq{\n\t}, $fm_projects->get_failed ), qq{\n},
    );

    return 1;

}

sub destroy {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dist_args_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        verbose  => { type => SCALAR, optional => 1 },
    });

    EFS::CLI::Namespace->assert_no_active_rebuilds;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{destroy_rpc},
        args   => [%args],
    );

}

sub destroy_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset(q{History});
    my $history = $history_rs->retrieve_assert( id => $args{history}, );

    my $project_dir = $project->get_path(q{dev});

    my $cell_rs = $class->get_resultset(q{Cell});

    my @dev_cells = $cell_rs->search({ celltype => q{dev} });

    my $metaproj = $project->metaproj;

    if ( my @releases = $class->sort_objects( $project->releases ) ) {
        die(
            qq{Unable to destroy project '$project', },
            qq{since it still has releases:\n\t},
            join ( qq{\n\t}, @releases ),
            qq{\n},
            q{All releases must be destroyed AND purged },
            qq{before the project can be destroyed.\n}
        );
    }

    my $cvs_dir = $project->get_path(q{dev}) . q{/CVS};

    if ( -d $cvs_dir ) {

        my $cvs = IO::Dir->new($cvs_dir) ||
            die qq{Unable to opendir $cvs_dir: $ERRNO\n};

        if ( grep { $_ ne q{.} && $_ ne q{..} } $cvs->read ) {
            die(
                qq{Unable to destroy project '$project', },
                qq{because the CVS directory is not empty.\n},
                q{Archive the contents, and purge the directory },
                qq{before destroying this project.\n}
            );
        }

    }

    my $project_lock;

    if ( -d $project_dir ) {
        $project_lock = EFS::DirLock->new({
            directory => $project_dir,
            alternate => qq{$project_dir/.destroyed},
            retry     => 1,
            command   => $history->request,
        });
    }

    my @dist_cells = (
        $cell_rs->search({ celltype => q{prod}, hidden => 0 }),
        $cell_rs->search({ celltype => q{dev},  hidden => 0 }),
    );

    my $fm_cells = EFS::ForkManager->new({

        list => \@dist_cells,

        action => sub {

            my ($cell) = @_;

            EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{destroy_cell},
                args   => [
                    cell     => $cell->name,
                    metaproj => $metaproj->name,
                    name     => $project->name,
                    type     => q{dist},
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to remove /efs/dist/$project from the following cells:\n\t},
        join ( qq{\n\t}, $fm_cells->get_failed ), qq{\n},
    );

    EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{destroy_cell},
        args   => [
            cell     => $project->cell->name,
            metaproj => $metaproj->name,
            name     => $project->name,
            type     => q{dev},
        ],
    );

    EFS::CLI::Namespace->delentry(
        type     => q{dev},
        cells    => \@dev_cells,
        realpath => $project->object_id,
        showpath => $project->get_path(q{dev}),
    );

    my $object_id = $project->object_id;

    $project->attrs->delete_all;
    $project->delete;

    STDOUT->print( qq{Destroyed project $object_id\n} );

    return 1;

}

sub destroy_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        cell     => { type => SCALAR },
        type     => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;
    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $cell_rs = $class->get_resultset(q{Cell});
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $type = $args{type};

    given ( $cell->fstype ) {
        
        when ( q{afs} ) {
            my $parentvolume = $project->metaproj->get_afsvolume( $type );
            my $afsvolume    = $project->get_afsvolume( $type );
            EFS::AFS::Util->remove_mount(
                volume => $afsvolume->name,
                parent => $parentvolume->name,
                path   => $project->name,
                cell   => $cell->name,
            );
            EFS::AFS::Util->remove_volume(
                name => $afsvolume->name,
                cell => $cell->name,
            );
        }

        when ( q{nfs} ) {

            my $admin_path = $cell->admin_path( type => $type );

            my @destroy = glob( qq{$admin_path/*/$type/$project} );

            if ( -d qq{$admin_path/$type/$project} ) {
                push @destroy, qq{$admin_path/$type/$project};
            }

            if ( not @destroy ) {
                STDOUT->print( qq{No directory $cell:/efs/$type/$project to destroy\n} );
                return 1;
            }

            EFS::Util->RemoveDirHierarchies( tops => \@destroy );

        }

    }

    return 1;

}

sub purge {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::validate_purge_args,
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{purge_rpc},
        args   => [%args],
    );

}

sub purge_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
        default_scope   => q{global},
    );

    my @releases = $project->destroyed_releases;

    if ( not @releases ) {
        STDOUT->print( qq{No destroyed releases in '$project' to purge.\n} );
    }

    STDOUT->print( qq{Purging destroyed releases from project '$project'.\n} );

    $args{project} = delete $args{name};

    my $fm_releases = EFS::ForkManager->new({

        parallel => 1,

        list => \@releases,

        action => sub {

            my ($release) = @_;

            my $object_id = $release->object_id(q{ });

            EFS::RPC->execute(
                class  => q{EFS::CLI::Release},
                method => q{purge_rpc},
                args   => [
                    authorized => 1,
                    name       => $release->name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{purge release $object_id},
                        metaproj => $project->metaproj->name,
                        project  => $project->name,
                        release  => $release->name,
                    },
                },
            );

            return 1;

        },

    });

    $fm_releases->run || die(
        qq{Unable to purge the following releases:\n\t},
        join ( qq{\n\t}, $fm_releases->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Successfully purged releases of project '$project'\n} );

    return 1;

}

sub chown {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        name      => { type => SCALAR },
        username  => { type => SCALAR, optional => 1 },
        norecurse => { type => BOOLEAN, optional => 1 },
        verbose   => { type => BOOLEAN, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{chown_rpc},
        args   => [%args],
    );

}

sub chown_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    STDOUT->print( qq{Chowning project '$project'\n} );

    my $metaproj  = $project->metaproj;
    my $groupname = $project->get_owner( type => q{group} );
    my $username  = $project->get_owner( type => q{user} );

    if ( $args{username} and $args{username} ne $username ) {

        my $username_obj = EFS::Username->new({ name => $args{username} });

        if ( not $username_obj->is_member_of( $groupname ) ) {
            die(
                qq{Invalid username '$args{username}'.\n},
                qq{Must be a member of the group '$groupname'\n},
            );
        }

        $username = $args{username};

    }

    my $lockgroup = $project->get_owner(
        lockmode => q{lock},
        type     => q{group},
    );

    # For backwards compatibility, we check for the existence of CVS
    # and RCS directories and chown them, too
    foreach my $subdir ( qw( CVS RCS ) ) {
        if ( not -d qq{/efs/dev/$project/$subdir} ) {
            next;
        }
        EFS::Util->Chown(
            path          => qq{/efs/dev/$project/$subdir},
            user          => $username,
            group         => $groupname,
            lockgroup     => $lockgroup,
            verbose       => $args{verbose},
            noworldaccess => $project->get_noworldaccess,
        );
    }

    if ( $args{norecurse} ) {
        return 1;
    }

    my @releases = $class->sort_objects( $project->releases );

    if ( not @releases ) {
        STDOUT->print( qq{No releases to chown in project '$project'\n} );
        return 1;
    }

    STDOUT->print( qq{Chowning all releases in project '$project'\n} );

    $args{project} = delete $args{name};

    my $fm_releases = EFS::ForkManager->new({

        list => \@releases,

        action => sub {

            my ($release) = @_;

            if ( $release->destroyed ) {
                return 1;
            }

            my $object_id = $release->object_id(q{ });

            EFS::RPC->execute(
                class  => q{EFS::CLI::Release},
                method => q{chown_rpc},
                args   => [
                    authorized => 1,
                    name       => $release->name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{chown release $object_id},
                        metaproj => $metaproj->name,
                        project  => $project->name,
                        release  => $release->name,
                    },
                },
            );

            return 1;

        },

    });

    $fm_releases->run || die(
        qq{Unable to chown the following releases:\n\t},
        join ( qq{\n\t}, $fm_releases->get_failed ), "\n"
    );

    STDOUT->print( qq{Successfully chowned releases of project '$project'\n} );

    return 1;

}

sub disttree {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my @releases       = $project->active_releases;
    my @releasealiases = $project->releasealiases;

    if ( grep { $_->celltype eq q{prod} } @cells ) {
        @releases       = grep { $_->has_only_prod_stages } @releases;
        @releasealiases = grep { $_->has_only_prod_stages } @releasealiases;
    }

    my $tree = EFS::DependencyTree->new({
        seeds => [ @releases, @releasealiases ],
        type  => q{runtime},
    });

    delete $args{metaproj};
    delete $args{name};

    if ( $args{dryrun} ) {
        STDOUT->print( $tree->get_tree_text( cells => \@cells ) );
        return 1;
    }
    else {
        return $tree->dist(%args);
    }

}

sub dist {
    return die(
        qq{This operation is no longer supported.\n},
        qq{Use 'efs disttree project' instead.\n},
    );
}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::validate_diststate_args,
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [%args],
    );

}

sub execlinks_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $metaproj = $project->metaproj;

    my @releases = $class->sort_objects( $project->releases );

    if ( not @releases ) {
        STDOUT->print(
            qq{There are no releases created yet for project '$project'\n},
        );
        return 1;
    }

    my @cells = $class->handle_cell_arguments(
        %args,
        default_scope => q{global},
    );

    STDOUT->print( qq{Updating execlinks for project '$project'\n} );

    $args{project} = delete $args{name};

    my $fm_releases = EFS::ForkManager->new({

        list => \@releases,

        action => sub {

            my ($release) = @_;

            if ( not $release->installs->count ) {
                return 1;
            }

            if ( $release->is_disabled ) {
                return 1;
            }

            if ( $release->destroyed ) {
                return 1;
            }

            my $object_id = $release->object_id(q{ });

            EFS::RPC->execute(
                class  => q{EFS::CLI::Release},
                method => q{execlinks_rpc},
                args   => [
                    authorized => 1,
                    name       => $release->name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{execlinks release $object_id},
                        metaproj => $metaproj->name,
                        project  => $project->name,
                        release  => $release->name,
                    },
                },
            );

            return 1;

        },

    });

    $fm_releases->run || die(
        qq{Unable to update execlinks for the following releases:\n\t},
        join ( qq{\n\t}, $fm_releases->get_failed ), qq{\n}
    );

    STDOUT->print(
        qq{Successfully updated all execlinks for project '$project'\n},
    );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        sortby   => { type => SCALAR, default => q{name}, optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    $class->display_objects( objects => [$project] );

    # XXX: ARGHH!!!  More column groups.
    my @optional = $project->get_efs_column_group(q{optional});

    my $feature_rs = $class->get_resultset(q{Feature});

    if ( not $feature_rs->configured(q{disable_restricted_access}) ) {
        push @optional, $project->get_efs_column_group(q{access});
    }

    if (@optional) {
        $project->display_options( columns => \@optional );
    }

    if ( my @attrs = $project->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( my @releasealiases = $project->releasealiases ) {
        $releasealiases[0]->display_objects( objects => \@releasealiases );
    }

    if ( my @releases = $project->releases ) {

        if ( my @active = grep { not $_->destroyed } @releases ) {
            $active[0]->display_objects(
                title       => q{Active Releases},
                objects     => \@active,
                sort_column => $args{sortby},
            );
        }

        if ( my @destroyed = grep { $_->destroyed } @releases ) {
            $destroyed[0]->display_objects(
                title           => q{Destroyed Releases},
                objects         => \@destroyed,
                display_columns => [
                    qw(metaproj project name stage locked
                        destroyed spaceused filecount
                        dircount linkcount)
                ],
                sort_column => $args{sortby},
            );
        }

    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{project} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub audit {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
        audits   => { type => SCALAR,
                      default => q{required},
                      optional => 1 },
        force    => { type => BOOLEAN,
                      optional => 1 },
        verbose  => { type => BOOLEAN,
                      optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{audit_rpc},
        args   => [%args],
    );

}

sub audit_rpc {

    my $class = shift;

    my %args = @_;

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $metaproj = $project->metaproj;

    my @releases = $class->sort_objects( $project->releases );

    if ( not @releases ) {
        STDOUT->print(
            qq{Unable to audit project '$project', },
            qq{since it has no active releases.\n}
        );
        return 1;
    }

    STDOUT->print(
        qq{Performing $args{audits} audits for all releases },
        qq{in project '$project'\n},
    );

    $args{project} = delete $args{name};

    my $fm_releases = EFS::ForkManager->new({

        list => \@releases,

        action => sub {

            my $release = shift;

            if ( $release->stage eq q{rejected} ) {
                STDOUT->print(
                    qq{Skipping audits for rejected release '$release'.\n}
                );
                return 1;
            }

            my $object_id = $release->object_id(q{ });

            EFS::RPC->execute(
                class  => q{EFS::CLI::Release},
                method => q{audit_rpc},
                args   => [
                    authorized => 1,
                    name       => $release->name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{audit release $object_id},
                        metaproj => $project->metaproj->name,
                        project  => $project->name,
                        release  => $release->name,
                    },
                },
            );

            return 1;

        },

    });

    $fm_releases->run || die(
        qq{Unable to successfully audit the following releases:\n\t},
        join ( qq{\n\t}, $fm_releases->get_failed ), qq{\n}
    );

    STDOUT->print(
        qq{Successfully audited all releases in project '$project'\n}
    );

    return 1;

}

sub set_distgroup {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        name      => { type => SCALAR },
        distgroup => { type => SCALAR },
    });

    my $feature_rs = $class->get_resultset(q{Feature});

    if ( $feature_rs->configured(q{disable_restricted_access}) ) {
        die(
            qq{Unable to set the distgroup attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $distgroup = EFS::Groupname->new({ name => $args{distgroup} });

    if ( $project->distgroup eq $distgroup->name ) {
        STDOUT->print(
            qq{Project $project already has distgroup $distgroup\n},
        );
    }
    else {
        $project->set_column( distgroup => $distgroup->name );
        $project->update;
        STDOUT->print(
            qq{Set distgroup for project $project to $distgroup\n},
        );
    }

    return 1;

}

sub clear_distgroup {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $feature_rs = $class->get_resultset(q{Feature});

    if ( $feature_rs->configured(q{disable_restricted_access}) ) {
        die(
            qq{Unable to set the distgroup attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    if ( $project->distgroup ) {
        $project->set_column( distgroup => undef );
        $project->update;
        STDOUT->print( qq{Cleared distgroup for project $project\n}, );
    }
    else {
        STDOUT->print( qq{Project $project has no distgroup\n} );
    }

    return 1;

}

sub set_noworldaccess {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj      => { type => SCALAR },
        name          => { type => SCALAR },
        noworldaccess => { type => SCALAR },
    });

    my $feature_rs = $class->get_resultset(q{Feature});

    if ( $feature_rs->configured(q{disable_restricted_access}) ) {
        die(
            qq{Unable to set the noworldaccess attribute, since the\n},
            qq{disable_restricted_access feature has been enabled.\n},
        );
    }

    my $resultset = $class->get_resultset;

    my $project = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{name},
    );

    my $noworldaccess = $args{noworldaccess} ? 1 : 0;

    $project->set_column( noworldaccess => $noworldaccess );
    $project->update;

    STDOUT->print(
        qq{Set noworldaccess for project $project to $noworldaccess\n},
    );

    return 1;

}

sub set_dist2dev {
    warn( qq{The set_dist2dev command has been deprecated\n} );
    return 1;
}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
