package EFS::CLI::ReleaseAlias;

use strict;
use warnings;
use English;

use feature q{switch};

use base q{EFS::CLI};
use base q{EFS::DBI};
use base q{EFS::DistableObject};

use Params::Validate qw( :all );

sub create {

    my $class           = shift;

    my (%args)          = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj        => { type       => SCALAR },
        project         => { type       => SCALAR },
        name            => { type       => SCALAR },
        target          => { type       => SCALAR },
        nocreate        => { type       => SCALAR,
                             optional   => 1 },
        _project_locked => { type       => SCALAR,
                             optional   => 1 },
    });

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{create_rpc},
        args   => [ %args ],
    );

}

sub create_rpc {

    my $class = shift;
    my %args  = @_;

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    my $project_dir = $project->get_path(q{dev});

    my $project_lock = $args{_project_locked} || EFS::DirLock->new({
        directory => $project_dir,
        retry     => 1,
        command   => $history->request,
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my ($release) = $release_rs->search({
        project => $project->id,
        name    => $args{name},
    });

    if ( $release ) {
        die(
            qq{Unable to create releasealias '$release' },
            qq{since it already exists as a release.\n}
        );
    }

    my $target = $release_rs->retrieve_release_or_alias_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{target},
    );

    my $target_stage = $target->stage;

    if ( $target_stage eq q{rejected} ) {
        die(
            qq{The target release '$target' is in stage '$target_stage'\n},
            qq{Release aliases may not resolve to $target_stage releases.\n}
        );
    }

    my $resultset = $class->get_resultset;

    my ($releasealias) = $resultset->search({
        project => $project->id,
        name    => $args{name},
    });

    my $verb;

    my $releasealias_lock;

    if ( $releasealias ) {

        $releasealias_lock = EFS::DirLock->new({
            directory => $project_dir,
            lockname  => qq{.lock-$args{name}},
            retry     => 1,
            command   => $history->request,
        });

        my @dependents = $releasealias->all_dependents( type => q{runtime} );

        if ( @dependents ) {

            my $releasealias_stage = $releasealias->stage;

            my @prod_dependents =
                grep { $_->dependent_release->has_prod_stages } @dependents;

            if ( $releasealias->has_prod_stages &&
                     $target_stage ne q{prod} &&
                         @prod_dependents ) {
                die(
                    q{The releasealias currently has prod dependents, and },
                    qq{resolves to stage '$releasealias_stage'\n},
                    qq{The new target is in stage '$target_stage'.\n},
                    q{Once a releasealias has prod dependents },
                    qq{you can not change it\n},
                    qq{to resolve to a '$target_stage' release.\n},
                );
            }

        }

        my $releasealias_id = $releasealias->object_id;

        foreach my $type ( qw( runtime build test ) ) {

            my $dependency_tree = EFS::DependencyTree->new({
                seeds => [ $target ],
                type  => $type,
            });

            if ( $dependency_tree->has_dependency( $releasealias_id ) ) {
                STDOUT->print(
                    qq{This would create a circular $type dependency:\n},
                    $dependency_tree->get_tree_text(
                        highlight => { $releasealias_id => q{circular} },
                    ),
                );
                die qq{Invalid releasealias target.\n};
            }

        }

        if ( $releasealias->target->name ne $args{target} ) {
            $releasealias->set_column( target => $args{target} );
            $releasealias->update;
            $verb = q{Updated};
        } else {
            $verb = q{Verified};
        }

    } else {

        $releasealias = $resultset->create({
            project => $project->id,
            name    => $args{name},
            target  => $args{target},
        });

        $verb = q{Created};

    }

    if ( $args{nocreate} ) {
        STDOUT->print( qq{Defined releasealias '$releasealias'\n} );
        return $releasealias;
    }

    EFS::Util->MakeLink(
        from => qq{/efs/dev/$releasealias},
        to   => $args{target},
    );

    STDOUT->print( qq{$verb releasealias $releasealias -> $args{target}\n} );

    return 1;

}

sub disttree {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $tree = EFS::DependencyTree->new({
        seeds => [ $releasealias ],
        type  => q{runtime},
    });

    delete $args{metaproj};
    delete $args{project};
    delete $args{name};

    if ( $args{dryrun} ) {
        STDOUT->print( $tree->get_tree_text( cells => \@cells ) );
        return 1;
    } else {
        return $tree->dist( %args );
    }

}

sub dist {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj        => { type       => SCALAR },
        project         => { type       => SCALAR },
        name            => { type       => SCALAR },
        _waitfor_cell   => { type       => SCALAR, optional   => 1 },
        _project_locked => { type       => SCALAR, optional   => 1 },
        _missing_cells_only
                        => { type       => SCALAR, optional   => 1 },
        q{no-strict-releasealiases}
                        => { type       => SCALAR, optional   => 1 },
        %EFS::CLI::validate_dist_args,
    });

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $releasealias->project->cell,
        class  => $class,
        method => q{dist_rpc},
        args   => [ %args ],
    );

}

sub dist_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project     = $releasealias->project;
    my $project_dir = $project->get_path(q{dev});

    my $target         = $releasealias->target;
    my $target_release = $releasealias->target_release;
    my $target_stage   = $target_release->stage;

    my $target_type =
        $target->isa(q{EFS::DBI::Result::Release}) ? q{release} : q{releasealias};

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    if ( not $target_release->installs->count ) {
        die(
            qq{Cannot dist releasealias because the target release has no\n},
            qq{installs, and hence no diststate.  Create installs for the\n},
            qq{target and dist the release before proceeding.\n}
        );
    }

    my $target_diststate = $target->get_diststate;

    if ( not $target_diststate->cells_partially_updated ) {
        die(
            qq{The target release(link) has not been disted anywhere yet.\n},
            qq{Dist the target object before disting this releasealias.\n},
        );
    }

    my $undisted = 0;
    my $partial  = 0;

    if ( $target_diststate->cells_needing_update( cells => \@cells ) ) {

        STDOUT->print( qq{Diststate for target $target_type $target:\n} );
        $target_diststate->display_text_table( cells => \@cells );

        if ( $target_type eq q{release} ) {
            $undisted++;
        } else {
            if ( $target_diststate->cells_never_updated( cells => \@cells ) ) {
                $undisted++;
            } else {
                $partial++;
            }
        }

    }

    if ( $undisted ) {
        die(
            qq{Unable to dist releasealias because the target:\n\t},
            qq{\t$target\n},
            qq{is not disted to all requested cells.  Dist the target to those\n},
            qq{cells before proceeding.\n},
        );
    }

    if ( $partial and not $args{q{no-strict-releasealiases}} ) {
        die(
            qq{Unable to dist releasealias because the target releasealias\n\t},
            qq{\t$target\n},
            qq{is not consistently deployed in all the specified cells.\n},
            qq{If this is acceptable, then you can dist is by using\n},
            qq{the --no-strict-releasealiases argument (--nsra is an alias)\n},
        );
    }

    my $project_lock = $args{_project_locked} || EFS::DirLock->new({
        directory => $project_dir,
        lockname  => q{.lock-} . $releasealias->name,
        retry     => 1,
        command   => $history->request,
    });

    my $diststate = $releasealias->get_diststate;

    if ( $args{_missing_cells_only} ) {
        @cells = $diststate->cells_never_updated( cells => \@cells );
    }

    if ( $diststate->needs_update( cells => \@cells ) ) {
        $diststate->display_text_table( cells => \@cells );
    } else {
        STDOUT->print(
            qq{Releasealias $releasealias is up to date in all requested cells.\n},
        );
        return 1;
    }

    return 1 if $args{dryrun};

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{dist_cell},
                args   => [
                    cell     => $cell->name,
                    metaproj => $releasealias->metaproj->name,
                    project  => $releasealias->project->name,
                    name     => $releasealias->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to distribute the releasealias to the following cells:\n\t},
        join(qq{\n\t}, map { $_->name } $fm_cells->get_failed ) . qq{\n}
    );

    STDOUT->print(
        qq{Successfully distributed releasealias $releasealias to all cells\n}
    );

    if ( $args{_waitfor_cell} ) {

        my $releasealias_dir = $releasealias->get_path(q{dist});

        if ( grep { $args{_waitfor_cell} eq $_ } $fm_cells->get_succeeded ) {
            STDOUT->print( qq{_EFS_WAITFOR: $releasealias_dir\n} );
        }

    }

    return 1;

}

sub dist_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell     => { type => SCALAR },
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $releasealias_rs = $class->get_resultset( q{ReleaseAlias} );
    my $releasealias = $releasealias_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $project = $releasealias->project;

    my $target      = $releasealias->target;
    my $target_name = $target->name;

    given ( $cell->fstype ) {
        when ( q{afs} ) {
            die qq{Unable to dist releasealiases in AFS cells yet\n};
            # XXX: This might be as trivial as just making a symlink, and
            # releasing a single volume
        }
        when ( q{nfs} ) {
            my $admin_path = $cell->admin_path( type => q{dist} );
            EFS::Util->MakeLink(
                from => qq{$admin_path/dist/$releasealias},
                to   => $target_name,
            );
        }
    }

    my $cellstate_rs = $class->get_resultset( q{ReleaseAliasCellState} );

    $cellstate_rs->update_or_insert(
        releasealias => $releasealias,
        cell         => $cell,
        target       => $target_name,
    );

    STDOUT->print( qq{Distributed releasealias $releasealias -> $target_name to $cell\n} );

    return 1;

}

sub destroy {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $releasealias->project->cell,
        class  => $class,
        method => q{destroy_rpc},
        args   => [ %args ],
    );

}

sub destroy_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my $project_dir = $releasealias->project->get_path(q{dev});

    my $project_lock = EFS::DirLock->new({
        directory => $project_dir,
        lockname  => q{.lock-} . $releasealias->name,
        retry     => 1,
        command   => $history->request,
    });

    if ( my @releasealiases = $releasealias->releasealiases ) {
        die(
            qq{The following releasealiases must be destroyed or changed before\n},
            qq{this releasealias can be destroyed:\n\t},
            join( qq{\n\t}, @releasealiases ), qq{\n}
        );
    }

    if ( $releasealias->releasealias_dependents ) {
        STDOUT->print( qq{The releasealias '$releasealias' has dependents defined.\n} );
        EFS::CLI::Dependents->display(
            metaproj => $releasealias->metaproj->name,
            project  => $releasealias->project->name,
            release  => $releasealias->name,
        );
        die(
            qq{Unable to destroy releasealias '$releasealias' until these },
            qq{dependencies have been undefined.\n}
        );
    }

    my $diststate = $releasealias->get_diststate;

    my @cells = $diststate->cells_partially_updated;

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{destroy_cell},
                args   => [
                    cell     => $cell->name,
                    metaproj => $releasealias->metaproj->name,
                    project  => $releasealias->project->name,
                    name     => $releasealias->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to destroy '$releasealias' in the following cells:\n\t},
        join(qq{\n\t}, map { $_->name } $fm_cells->get_failed ) . qq{\n}
    );

    # XXX: This should work for BOTH nfs and afs
    if ( -l qq{/efs/dev/$releasealias} ) {
        unlink( qq{/efs/dev/$releasealias} ) ||
            die qq{Unable to unlink /efs/dev/$releasealias: $ERRNO\n};
        STDOUT->print( qq{Destroyed '$releasealias' in /efs/dev.\n} );
    }

    my $object_id = $releasealias->object_id;

    $releasealias->cellstates->delete_all;
    $releasealias->attrs->delete_all;
    $releasealias->delete;

    STDOUT->print( qq{Successfully destroyed releasealias $object_id in all cells\n} );

    return 1;

}

sub destroy_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell     => { type => SCALAR },
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $releasealias_rs = $class->get_resultset( q{ReleaseAlias} );
    my $releasealias = $releasealias_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    my $project     = $releasealias->project;
    my $target      = $releasealias->target;
    my $target_name = $target->name;

    given ( $cell->fstype ) {
        when ( q{afs} ) {
            die qq{Unable to destroy releasealiases in AFS cells yet\n};
        }
        when ( q{nfs} ) {
            my $admin_path = $cell->admin_path( type => q{dist} );

            my $dist_link = qq{$admin_path/dist/$releasealias};

            if ( -l $dist_link ) {
                unlink( $dist_link ) ||
                    die qq{Unable to unlink $dist_link: $ERRNO\n};
                STDOUT->print(
                    qq{Destroyed releasealias $releasealias -> $target_name in cell $cell\n}
                );
            } else {
                # NOTE: ReleaseAlias diststate may change this.
                STDOUT->print(
                    qq{Releasealias $releasealias was never disted to $cell.  },
                    qq{Nothing to destroy.\n},
                );
            }
        }
    }

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR | OBJECT },
        project  => { type => SCALAR | OBJECT },
        name     => { type => SCALAR | OBJECT },
    });

    my $resultset = $class->get_resultset;

    my $releasealias = $resultset->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    $class->display_objects( objects => [$releasealias] );

    if ( my @attrs = $releasealias->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    if ( $releasealias->all_dependents ) {
        EFS::CLI::Dependency->display(
            metaproj => $releasealias->metaproj->name,
            project  => $releasealias->project->name,
            release  => $releasealias->name,
        );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{releasealias} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
