package EFS::DBI;

use strict;
use warnings;
use English;

use base q{DBIx::Class::Schema};

if ( -f q{/etc/efs/server.conf} ) {
    __PACKAGE__->load_namespaces;
} else {
    warn qq{WARNING: EFS API database access only supported on EFS servers\n};
}

use feature q{switch};

# Pure OO modules, loaded ONCE, listed here.
use IO::Dir;
use Text::FormatTable;

# modules used in THIS file...
use Readonly;
use Params::Validate qw( :all );
use Regexp::Common qw(number);
use Lingua::EN::Inflect ();
use Data::Dumper;
use Date::Format;
use Date::Calc qw(:all);
use Date::Manip;
use SQL::Abstract;
use List::MoreUtils qw( uniq );
use Try::Tiny;

use EFS::Config::Server;

Readonly my $SCREEN_WIDTH => 1024;

sub get_subclass {
    my $proto = shift;
    my $class = ref $proto || $proto;
    my @parts = reverse split m/::/msx, $class;
    return $parts[0];
}

sub get_resultset {
    my $class    = shift;
    my $subclass = shift || $class->get_subclass;
    my $schema   = $class->get_schema;
    return $schema->resultset($subclass);
}

sub get_columns {
    my $class = shift;
    my $source = shift;
    my $schema = $class->get_schema;
    return $schema->source($source)->columns;
}

sub retrieve_bynames_assert {
    my $class     = shift;
    my $resultset = $class->get_resultset( $class->get_subclass );
    return $resultset->retrieve_bynames_assert(@_);
}

sub get_localcell {
    my $class     = shift;
    my $config    = EFS::Config::Client->new;
    my $resultset = $class->get_resultset( q{Cell} );
    return $resultset->retrieve_bynames_assert(
        name => $config->get_value( q{client}, q{cell} )
    );
}

#
# NOTE: This is a bit ugly, and exposes some assumptions in the code. This is
# schema specific, NOT class specific, and only used when constructing raw SQL
# in EFS::CLI::Database->dumpconfig_file().
#
# The code assumes there is ONE database, so when we're working with both a
# real RDBMS and SQLite, we have to fake some things.
#
sub get_table_string {
    my $schema = shift;
    my $source = shift;
    my $quote  = shift;
    my $separator = shift || q{.};
    my $table = $schema->source( $source )->from;
    $table =~ s{\Q$separator\E}{$quote$separator$quote}msx;
    return qq{$quote$table$quote};
}

{

    my %subclass2type = (
        AFSVolume              => q{afsvolume},
        AutoRelease            => q{autorelease},
        ChildRelease           => q{childrelease},
        DependencyRelease      => q{dependency_release},
        DependencyReleaseAlias => q{dependency_releasealias},
        DistStats              => q{diststats},
        EFSServer              => q{efsserver},
        EFSServerCell          => q{efsservercell},
        ForbiddenFileRule      => q{ffrule},
        GroupRights            => q{grouprights},
        InstallAuditState      => q{auditstate},
        PTS                    => q{pts},
        PlatformBuild          => q{platformbuild},
        PlatformInstance       => q{platforminstance},
        PlatformAlias          => q{platformalias},
        PlatformInstanceAlias  => q{platforminstancealias},
        ReleaseAlias           => q{releasealias},
        SSLServer              => q{sslserver},
        UserGroup              => q{usergroup},
        UserRights             => q{userrights},
        # NOTE: This one can go away in EFS 3.1
        ReleaseLink            => q{releaselink},
    );

    my %type2subclass = reverse %subclass2type;

    sub type2subclass {
        my ( $class, $type ) = @_;
        return $type2subclass{$type} || ucfirst ($type);
    }

    sub class2type {

        my $class = shift;
        my $classname = shift || ref $class || $class;

        my ($subclass) = reverse split m/:+/msx, $classname;

        if ( $subclass =~ m/Attrs \z/msx ) {
            return q{attribute};
        }
        elsif ( $subclass2type{$subclass} ) {
            return $subclass2type{$subclass};
        }
        else {
            return lc ($subclass);
        }

    }

}

sub type2plural {
    my ( $class, $type ) = @_;
    return Lingua::EN::Inflect::PL($type);
}

sub cli_order {

    my $class = shift;
    my %args  = @_;

    my $type   = $args{type}   || $class->class2type;
    my $action = $args{action} || q{default};

    if ( $class =~ s/Attrs \z//gmsx ) {
        return EFS::CLI->cli_order(
            parent_type => $class->class2type,
            type        => $type,
            action      => $action,
        );
    }
    else {
        return EFS::CLI->cli_order( type => $type, action => $action, );
    }

}

sub attrsclass {
    my $proto = shift;
    my $class = ref $proto || $proto;
    return qq{${class}Attrs};
}

sub attrssubclass {
    my $proto = shift;
    my $class = ref $proto || $proto;
    my ($subclass) = reverse split m/::/msx, $class;
    return qq{${subclass}Attrs};
}

sub is_disabled {

    my $self = shift;

    my $disabled = $self->get_attribute( name => q{disabled} );

    if ( $disabled && $disabled->value ) {
        return $disabled->value;
    }
    else {
        return;
    }

}

sub _sum_column {

    my $self   = shift;
    my $column = shift;

    if ( $self->can( q{installs} ) ) {
        my ($object) = $self->get_resultset( q{Install} )->search(
            { id => [ map { $_->id } $self->installs ], },
            { select => { sum => $column }, as => $column, },
        );
        if ( ref $object ) {
            my $result = $object->get_column($column);

            if (not defined $result) {
                # SQL summation rules mean that if there are no installs,
                # we'll get NULL--return zero instead
                return 0;
            }
            else {
                return $result;
            }
        }
        else {
            return 0;
        }
    }
    else {

        # Fatal exception, maybe?  If it can't call installs, this should't
        # have called this....
        return 0;
    }

}

sub spaceused {
    return shift->_sum_column(q{spaceused});
}

sub filecount {
    return shift->_sum_column(q{filecount});
}

sub dircount {
    return shift->_sum_column(q{dircount});
}

sub linkcount {
    return shift->_sum_column(q{linkcount});
}

sub size_shorthand {

    ## no critic (MagicNumbers)

    my $class = shift;
    my $size = shift || 0;

    return qq{$size bytes} if $size < 1000;

    $size /= 1000;
    $size = sprintf ( q{%.1f}, $size );
    $size = int ($size) if $size > 10;
    return qq{$size KB} if $size < 1000;

    $size /= 1000;
    $size = sprintf ( q{%.1f}, $size );
    $size = int ($size) if $size > 10;
    return qq{$size MB} if $size < 1000;

    $size /= 1000;
    $size = sprintf ( q{%.1f}, $size );
    $size = int ($size) if $size > 10;
    return qq{$size GB} if $size < 1000;

    $size /= 1000;
    $size = sprintf ( q{%.1f}, $size );
    $size = int ($size) if $size > 10;
    return qq{$size TB};

}

sub delta_days {

    my $class = shift;
    my $date = shift;

    $date =~ s/\s+ [^\n]* \z//msx;

    my ($now_year,$now_month,$now_day) = Today;

    my ($purge_year,$purge_month,$purge_day) =
        split( m/-/msx, $date) or
            die qq{Unable to parse date '$date'\n};

    return Delta_Days( $purge_year,$purge_month,$purge_day,
                       $now_year,$now_month,$now_day );

}

sub authorize {

    my $class = shift;
    my %args  = @_;

    my %authgroup = ();
    my %authuser  = ();

    if ( not $args{noefsops} ) {
        $authgroup{efsops}++;
    }

    if ( $args{authuser} eq q{nobody} ) {
        die(
            qq{The efs client is being run as root, therefore\n},
            qq{only commands which require no authorization are available.\n},
        );
    }

    my %effective_role = (
        rejected   => q{dev},
        deprecated => q{prod},
    );

    my $role = q{dev};

    my $object_id = q{};

    if ( $args{object} ) {

        $role = $args{object}->stage;
        $object_id = $args{object}->object_id;

    }
    else {

        if ( $args{type} eq q{autorelease} ) {

            $role = q{auto};

        }
        elsif ( $args{type} eq q{releasealias} ) {

            my $name = $args{action} eq q{create} ? q{target} : q{releasealias};

            my $resultset = $class->get_resultset(q{Release});

            my $object = $resultset->retrieve_release_or_alias_assert(
                metaproj => $args{history}->{metaproj},
                project  => $args{history}->{project},
                name     => $args{history}->{$name},
            );

            $role = $object->stage;

            $object_id = $object->object_id;

        } elsif ( $args{type} eq q{release} ) {

            my $resultset = $class->get_resultset(q{Release});

            my $release = $resultset->retrieve_bynames(
                metaproj => $args{history}->{metaproj},
                project  => $args{history}->{project},
                name     => $args{history}->{release},
            );

            if ($release) {
                $role = $release->stage;
                $object_id = $release->object_id;
            }

        }

    }

    $role = $effective_role{$role} || $role;

    my $efsops_only = 0;

    ##  no critic: CascadingIfElse
    if ( $args{action} ~~ [qw( create define )] &&
             $args{type} eq q{metaproj} ) {
        $efsops_only = 1;
    }
    elsif ( $args{action} ~~ [ qw( set_movelock clear_movelock ) ] ) {
        $efsops_only = 1;
    }
    elsif ( $args{action} eq q{purge} && $args{type} ne q{release} ) {
        $efsops_only = 1;
    }
    elsif (
        $args{action} ~~ [
            qw( set_owner_list set_notify_list
                test_owner_list test_notify_list )
        ]
        ) {
        $role = q{admin};
    }

    if ( $args{type} eq q{grouprights} || $args{type} eq q{userrights} ) {
        $role = q{admin};
    }

    if ( $efsops_only == 0 && $args{history}->{metaproj} ) {

        my $resultset = $class->get_resultset(q{Metaproj});

        my $metaproj = $resultset->retrieve_bynames_assert(
            name => $args{history}->{metaproj}, );

        if ( $role eq q{dev} ) {
            $authgroup{ $metaproj->groupname }++;
        }

        my @grouprights = $metaproj->grouprights({ role => $role });
        my @userrights = $metaproj->userrights({ role => $role });

        if ( $role eq q{qa} and not @grouprights and not @userrights ) {
            @grouprights = $metaproj->grouprights({ role => q{dev} });
            @userrights = $metaproj->userrights({ role => q{dev} });
            $authgroup{ $metaproj->groupname }++;
        }

        foreach my $grouprights (@grouprights) {
            $authgroup{ $grouprights->groupname }++;
        }

        foreach my $userrights (@userrights) {
            $authuser{ $userrights->username }++;
        }

    }

    my $username = EFS::Username->new({ name => $args{authuser} });

    my @groups = $username->groups;

    foreach my $authgroup ( sort keys %authgroup ) {
        if ( grep { $_ eq $authgroup } @groups ) {
            return 1;
        }
    }

    foreach my $authuser ( sort keys %authuser ) {
        if ( $args{authuser} eq $authuser ) {
            return 1;
        }
    }

    my $fatal = q{};

    if ( $args{action} ne q{disttree} or not $args{object} ) {
        $fatal = qq{Access denied.\n};
    }

    if ( not $efsops_only ) {
        if ( $object_id ) {
            $fatal .=
                qq{The '$role' role is required to $args{action} '$object_id'\n};
        } else {
            $fatal .=
                qq{The '$role' role is required to $args{action} $args{type}\n};
        }
    }

    return die($fatal);

}

sub get_noworldaccess {

    my $self = shift;

    my $resultset = $self->get_resultset(q{Feature});

    if ( $resultset->configured(q{disable_restricted_access}) ) {
        return 0;
    }
    else {
        return $self->find_mpr_attr(q{noworldaccess});
    }

}

sub get_dirmode {
    ## no critic (MagicNumbers)
    return shift->get_noworldaccess ? 02770 : 02775;
}

sub get_owner {

    my $self = shift;

    my (%args) = validate(
        @_,
        {   lockmode =>
                { type => SCALAR, optional => 1, default => q{unlock}, },
            type => { type => SCALAR },
        }
    );

    my $metaproj
        = $self->isa(q{EFS::DBI::Result::Metaproj}) ? $self : $self->metaproj;

    if ( $args{lockmode} eq q{unlock} ) {
        return ( $args{type} eq q{user}
            ? $metaproj->username
            : $metaproj->groupname );
    }

    if ( $args{lockmode} eq q{lock} && $args{type} eq q{user} ) {
        return q{root};
    }

    my $noworldaccess = $self->get_noworldaccess;

    if ( not $noworldaccess ) {
        return q{root};
    }

    if ( $self->isa(q{EFS::DBI::Result::Install}) ) {
        if ( my $distgroup = $self->find_mpr_attr(q{distgroup}) ) {
            if ( $args{type} eq q{user} ) {
                EFS::Username->new( { name => $distgroup } );
            }
            else {
                EFS::Groupname->new( { name => $distgroup } );
            }
            return $distgroup;
        }
        else {
            return $metaproj->groupname;
        }
    }
    else {
        return $metaproj->groupname;
    }

}

sub find_mpr_attr {

    my $self     = shift;
    my $attrname = shift;

    my @objects = ();

    ##  no critic: CascadingIfElse
    if ( $self->isa( q{EFS::DBI::Result::Install} ) ) {
        push ( @objects, $self->release, $self->project, $self->metaproj );
    }
    elsif ($self->isa( q{EFS::DBI::Result::Release} )
        || $self->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {
        push ( @objects, $self, $self->project, $self->metaproj );
    }
    elsif ( $self->isa( q{EFS::DBI::Result::Project} ) ) {
        push ( @objects, $self, $self->metaproj );
    }
    elsif ( $self->isa( q{EFS::DBI::Result::Metaproj} ) ) {
        push ( @objects, $self );
    }
    else {
        die(
            q{Invalid class: }, ref $self, qq{\n},
            qq{Must be one of:\n},
            qq{\tEFS::DBI::Result::Metaproj\n},
            qq{\tEFS::DBI::Result::Project\n},
            qq{\tEFS::DBI::Result::Release\n},
            qq{\tEFS::DBI::Result::Install\n},
        );
    }

    my $attribute = q{};

    foreach my $object (@objects) {

        if ( $object->can($attrname) ) {
            if ( $attribute = $object->$attrname ) {
                return $attribute;
            }
        }
        else {
            if ( $attribute = $object->get_attribute( name => $attrname ) ) {
                return $attribute->value;
            }
        }

    }

    if ( not $attribute ) {
        my $parameter_rs = $self->get_resultset( q{Parameter} );
        $attribute = $parameter_rs->configured( $attrname );
    }

    return $attribute;

}

sub handle_cell_arguments {

    my $class = shift;

    # Let's not bother validating anything here.  We're just passing
    # ALL of %args into this subroutine from the caller, so the
    # critical arguments have already been validated.
    my (%args) = @_;

    if ( exists $args{default_scope} and $args{default_scope} eq q{global} ) {
        if ( not $args{cells} and not $args{region} and
                 not $args{campus} and not $args{location} ) {
            $args{global} = 1;
        }
    }

    # This is a bit of a hack, I suppose...  Sometimes, an ARRAYREF of
    # cells is passed, so in that case, we just pass it back.  Why?
    # Well, EFS::DBI::Metaproj->dist handles all these args, and then
    # calls EFS::DBI::Project->dist, and if you already have expanded
    # the cells, we don't do it again.

    # Another hack...  If the array is a bunch of strings, convert them to
    # cell objects.  This is done to make the disttree implementation happy,
    # 'cause it wants to pass the list of cells across EFS::RPC, but we can't
    # do that with dbix objects

    my $cell_rs     = $class->get_resultset( q{Cell} );

    if ( ref $args{cells} eq q{ARRAY} ) {
        return map {
            ref $_ ? $_ : $cell_rs->retrieve_bynames_assert( name => $_ )
        } @{ $args{cells} };
    }

    # Exit if attempt made to use -cells with other cell spec args
    if ( exists $args{cells} ) {
        foreach my $spec ( qw(location campus region global) ) {
            if ( exists $args{$spec} ) {
                die(
                    qq{Mutually exclusive arguments: --cells can not be used with any of:\n},
                    qq{\t--location, --campus, --region or --global\n},
                );
            }
        }
    }

    # Namespace and Volume create don't support celltype
    # This avoids a warning later on...
    $args{celltype} ||= q{dev};

    my @cells = ();

    # When an explicit cell or hiddencells is passed in, we
    # do not limit the searches to those where hidden is 0.
    my $need_hidden = ( $args{cells} || $args{hiddencells} );

    my @hiddencells = ( hidden => 0 );
    if ($need_hidden) {
        @hiddencells = ();
    }

    my @candidates = ();

    if ( $args{global} ) {

        if ( $args{celltype} eq q{all} ) {
            # Huh?  Why just dev and prod?  We want to hide the test stuff by
            # default, unless you specifically ask for it by name
            @cells = (
                $cell_rs->search({ celltype => q{prod}, @hiddencells }),
                $cell_rs->search({ celltype => q{dev},  @hiddencells }),
            );
        } else {
            @cells = $cell_rs->search({ celltype => $args{celltype}, @hiddencells });
        }

        if ( $args{include_parents} ) {
            foreach my $cell ( @cells ) {
                next if not $cell->celltype ~~ [qw( test prod )];
                push @cells, $cell->devcell;
            }
        }

    } elsif ( $args{region} || $args{campus} || $args{location} ) {

        if ( $args{region} ) {
            my $region_rs = $class->get_resultset( q{Region} );
            foreach my $name ( split m{[,\h]+}msx, $args{region} ) {
                push @candidates,
                    $region_rs->retrieve_bynames_assert( name => $name )->cells;
            }
        }

        if ( $args{campus} ) {
            my $campus_rs = $class->get_resultset( q{Campus} );
            foreach my $name ( split m{[,\h]+}msx, $args{campus} ) {
                push @candidates,
                    $campus_rs->retrieve_bynames_assert( name => $name )->cells;
            }
        }

        if ( $args{location} ) {
            my $location_rs = $class->get_resultset( q{Location} );
            foreach my $name ( split m{[,\h]+}msx, $args{location} ) {
                push @candidates,
                    $location_rs->retrieve_bynames_assert( name => $name )->cells;
            }
        }

        # Filter out hidden cells here so we don't follow them
        if ( not $need_hidden ) {
            @candidates = grep { not $_->hidden } @candidates;
        }

        if ( $args{celltype} eq q{all} ) {
            @cells = @candidates;
        } else {
            foreach my $candidate ( @candidates ) {
                next if $candidate->celltype ne $args{celltype};
                push @cells, $candidate;
                next if not $args{include_parents};
                next if not $candidate->celltype ~~ [qw( test prod )];
                push @cells, $candidate->devcell;
            }
        }

        # Cells are NOT filtered using the celltype

    } elsif ( $args{cells} ) {

        foreach my $cell_name ( split( m/,/msx, $args{cells} ) ) {
            my $cell = $cell_rs->retrieve_bynames_assert( name => $cell_name );
            push @cells, $cell;
            next if not $args{include_parents};
            next if not $cell->celltype ~~ [qw( test prod )];
            push @cells, $cell->devcell;
        }

    } else {

        # This will be the localcell of the EFS server, NOT the client
        @cells = ( $class->get_localcell );

        if ( $args{celltype} ~~ [qw( prod all )] ) {
            # We know there's only one, and we skip test by default
            push @cells, grep { $_->celltype eq q{prod} } $cells[0]->children;
        }

    }

    if ( not @cells ) {
        die qq{Cell specification does not match any cells.\n};
    }

    # Sort cells as follows:
    # If -cell is explicitly given, return in the order requested
    # This is already handled by logic above
    if ( not $args{cells} ) {
        # Otherwise, sort using region/campus/location/etc.
        @cells = $class->sort_objects(@cells);
    }

    return uniq @cells;

}

sub sort_objects {

    my $class   = shift;
    my @objects = @_;

    my $sortsub = sub {
        if ( $a->can(q{sort_id}) ) {
            return lc($a->sort_id) cmp lc($b->sort_id);
        }
        elsif ( $a->can(q{object_id}) ) {
            return lc($a->object_id) cmp lc($b->object_id);
        }
        else {
            return lc($a) cmp lc($b);
        }
    };

    my @sorted_objs = sort $sortsub @objects;
    return @sorted_objs;

}

sub dependency_resultset {

    my $self = shift;

    if ( $self->isa( q{EFS::DBI::Result::Release} ) ) {
        return $self->get_resultset( q{DependencyRelease} );
    }
    elsif ( $self->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {
        return $self->get_resultset( q{DependencyReleaseAlias} );
    }
    else {
        return die(
            qq{Not a release or releasealias object:\n},
            qq{$self isa }, ref $self, qq{\n},
        );
    }

}

sub _attribute_data {

    my $class     = shift;
    my $operation = shift;

    my %args = @_;

    my $type = $class->class2type;

    my %method_of = (
        setattr => q{set_attribute},
        getattr => q{get_attribute},
        delattr => q{delete_attribute},
    );

    my $data = {

        method => $method_of{$operation},

        validate => {
            %EFS::CLI::validate_common,
            name => { type => SCALAR },
        },

        arguments => {
            name    => $args{name},
            verbose => 1,
        },

    };

    if ( $operation eq q{setattr} ) {
        $data->{validate}->{value} = { type => SCALAR };
        $data->{arguments}->{value} = $args{value};
    }

    foreach my $column ( EFS::CLI->cli_order( type => $type ) ) {

        next if $column eq q{id};

        if ( $column eq q{name} ) {
            $data->{validate}->{$type} = { type => SCALAR };
            $data->{retrieve}->{name} = $args{$type};
        }
        else {
            $data->{validate}->{$column} = { type => SCALAR };
            $data->{retrieve}->{$column} = $args{$column};
        }

    }

    return $data;

}

sub _attribute_ops {

    my $class     = shift;
    my $operation = shift;

    my $data = $class->_attribute_data( $operation, @_ );

    my (%args) = validate( @_, $data->{validate} );

    my $object = $class->retrieve_bynames_assert( %{ $data->{retrieve} } );

    my $method = $data->{method};

    return $object->$method( %{ $data->{arguments} } );

}

sub setattr {
    return shift->_attribute_ops( q{setattr}, @_ );
}

sub getattr {
    return shift->_attribute_ops( q{getattr}, @_ );
}

sub delattr {
    return shift->_attribute_ops( q{delattr}, @_ );
}

sub set_attribute {

    my $self = shift;

    my (%args) = validate( @_, {
        name    => { type => SCALAR },
        value   => { type => SCALAR },
        verbose => { type => SCALAR, optional => 1 },
    });

    my $class         = ref $self;
    my $object_id     = $self->object_id;
    my $type          = $class->class2type;
    my $attrssubclass = $self->attrssubclass;

    my $attrs_rs = $class->get_resultset($attrssubclass);

    my $attribute = $self->get_attribute( name => $args{name} );

    my $action = q{existing attribute is already};

    if ( $attribute ) {
        if ( $attribute->value ne $args{value} ) {
            $attribute->set_column( value => $args{value} );
            $attribute->update;
            $action = q{updated attribute to};
        }
    }
    else {
        $attribute = $attrs_rs->create({
            $type => $self->id,
            name  => $args{name},
            value => $args{value},
        });
        $action = q{defined attribute as};
    }

    if ( $args{verbose} ) {
        STDOUT->print(
            qq{For $type '$object_id', $action:\n},
            qq{\t'$args{name}' => '$args{value}'\n},
        );
    }

    return $attribute;

}

sub get_attribute {

    my $self = shift;

    my (%args) = validate( @_, {
        name    => { type => SCALAR },
        verbose => { type => SCALAR, optional => 1 },
    });

    my $class         = ref $self;
    my $object_id     = $self->object_id;
    my $type          = $class->class2type;
    my $attrsclass    = $self->attrsclass;
    my $attrssubclass = $self->attrssubclass;

    my $attrs_rs = $class->get_resultset($attrssubclass);

    my ($attribute) = $attrs_rs->search({
        $type => $self->id,
        name  => $args{name},
    });

    if ( $args{verbose} ) {
        if ($attribute) {
            $attrsclass->display_objects( objects => [$attribute] );
        }
        else {
            die qq{For $type '$object_id', no such attribute '$args{name}'\n};
        }
    }

    return $attribute;

}

sub delete_attribute {

    my $self = shift;

    my (%args) = validate( @_, {
        name    => { type => SCALAR },
        verbose => { type => SCALAR, optional => 1 },
    });

    my $class     = ref $self;
    my $object_id = $self->object_id;
    my $type      = $class->class2type;
    my $attribute = $self->get_attribute( name => $args{name} );

    if ($attribute) {
        $attribute->delete;
        if ( $args{verbose} ) {
            STDOUT->print(
                qq{For $type '$object_id', deleted attribute '$args{name}'\n},
            );
        }
    }
    else {
        if ( $args{verbose} ) {
            die qq{For $type '$object_id', no such attribute '$args{name}'\n};
        }
    }

    return 1;

}

sub display_headers {

    my $self = shift;
    my $display_columns = shift || [ $self->display_columns ];

    my $type = $self->class2type;

    my @headers = ();

    foreach my $column (@{ $display_columns }) {
        push ( @headers, ucfirst ( $column eq q{name} ? $type : $column ) );
    }

    return @headers;

}

sub display_columns {
    my $self = shift;
    return grep { $_ ne q{id} } $self->columns;
}

sub display_objects {

    my $class = shift;

    my (%args) = validate( @_, {
        objects         => { type => ARRAYREF },
        display_columns => { type => ARRAYREF, optional => 1 },
        sort_column     => { type => SCALAR,   optional => 1 },
        title           => { type => SCALAR,   optional => 1 },
    });

    my $type = $class->class2type;

    my @objects = @{ $args{objects} };

    if ( not @objects ) {
        STDOUT->print( qq{No objects to display\n} );
        return 1;
    }

    my @display_columns = ();

    if ( $args{display_columns} ) {
        @display_columns = @{ $args{display_columns} };
    }
    else {
        @display_columns = $objects[0]->display_columns;
    }

    my @display_headers = $objects[0]->display_headers( \@display_columns );

    my @table_format = map {q{r}} @display_columns;

    my $sort_column = $args{sort_column} || $display_columns[0];

    my $sort_operator = q{numeric};

    foreach my $object (@objects) {
        foreach my $index ( reverse 0 .. $#display_columns ) {

            my $column_name = $display_columns[$index];
            my $string      = $object->$column_name;

            if ( ref $string ) {
                $string = $string->name;
            }

            if ( $string && $string !~ m/\A $RE{num}{real} \z/msx ) {
                $table_format[$index] = q{l};
                if ( $column_name eq $sort_column ) {
                    $sort_operator = q{string};
                }
            }

        }
    }

    if ( $sort_operator eq q{string} ) {
        @objects =
            sort { lc($a->$sort_column) cmp lc($b->$sort_column) } @objects;
    }
    else {
        @objects =
            reverse sort { $a->$sort_column <=> $b->$sort_column } @objects;
    }

    my $table = Text::FormatTable->new( join(q{  },@table_format) );
    $table->head( @display_headers );
    $table->rule(q{=});

    foreach my $object (@objects) {

        my @row = ();

        my $object_class = ref $object;

        foreach my $column ( reverse @display_columns ) {

            my $value =
                $object->display_translate( $column => $object->$column );

            my $value_display
                = ref $value
                ? $value->name
                : ( defined $value ? $value : q{} );

            push ( @row, $value_display );

        }

        $table->row( map { ( defined($_) && $_ ne q{} ) ? $_ : q{NULL} }
                         reverse @row );

    }

    STDOUT->print( qq{\n} );

    if ( $args{title} ) {
        STDOUT->print( qq{$args{title}:\n} );
    }

    STDOUT->print($table->render($SCREEN_WIDTH));

    return 1;

}

sub get_efs_column_group {

    my $self         = shift;
    my $column_group = shift;

    my $result_source = $self->result_source;

    my @columns = ();

    foreach my $column ( $result_source->columns ) {
        my $column_info = $result_source->column_info($column);
        if ( ref $column_info->{efs_column_group} eq q{ARRAY} ) {
            if ( grep { $_ eq $column_group }
                @{ $column_info->{efs_column_group} } ) {
                push @columns, $column;
            }
        }
    }

    return @columns;

}

sub display_options {

    my $self = shift;

    my (%args) = validate( @_, {
        columns => {
            type     => ARRAYREF,
            default  => [ $self->get_efs_column_group(q{optional}) ],
            optional => 1
        },
    });

    my %options;

    foreach my $key ( @{ $args{columns} } ) {
        my $value = $self->$key;
        if ( defined $value ) {
            $options{$key} = $value;
        }
    }

    if ( not %options ) {
        return 1;
    }

    my $table = Text::FormatTable->new( q{l  r} );

    $table->head( qw(Option Value) );
    $table->rule;

    foreach my $key ( sort keys %options ) {
        $table->row( $key, $options{$key} );
    }

    STDOUT->print( qq{\n} );

    STDOUT->print( $table->render($SCREEN_WIDTH) );

    return 1;

}

sub assert_maintenance_mode {

    my $class = shift;

    my $parameter_rs = $class->get_resultset(q{Parameter});

    if ( not $parameter_rs->configured(q{maintenance_mode}) ) {
        die qq{This operation can only be run when EFS is in maintenance mode\n};
    }

    return 1;

}

sub assert_no_active_dists {

    my $class = shift;

    my $history_rs = $class->get_resultset( q{History} );
    my @active = $history_rs->search({
        status => undef,
        -or => [
            command => { like => q{dist %} },
            command => { like => q{disttree %} },
        ]
    });

    if ( @active ) {

        STDOUT->print(
            qq{The following dist commands are still active:\n},
        );

        $active[0]->display_objects(
            objects         => \@active,
            display_columns => [qw(id authuser efshost pid request)],
        );

        die(
            qq{This command may not be used while active dists are in progress\n},
        );

    }

    return 1;

}

sub admin_path_show {

    my $class = shift;
    my $dir = shift;

    if ( $dir !~ m{ ^ /.efsadmin/ }msx ) {
        return $dir;
    }

    ## no critic ComplexRegexes
    my ($admin_path,$cell,$type,$remainder) =
        ( $dir =~ m{
                       (
                           ^
                           /.efsadmin/
                           ( [^/]+ ) / # cell
                           [^/]+ /     # fileserver
                           .*/efs\d?_ ( \w+? ) (?:_\w+_\w+)? /efs_qtree
                       ) # admin_path
                       (?:
                           /
                           (
                               .*
                           ) # remainder
                       )?
               }msx );

    if ( not $admin_path or not $cell or not $type ) {
        die qq{Unable to parse admin path: $dir\n};
    }

    my $dir_show = qq{$cell:};

    my $efs = q{efs};

    if ( $remainder ) {

        my @dirs = split m{/}msx, $remainder;

        if ( $dirs[0] eq q{global} and $#dirs > 0 ) {
            $dirs[0] = $type;
        }

        if ( $dirs[0] =~ m{\.}msx ) {
            my $instance = shift @dirs;
            $efs = qq{.efs/$instance};
        }

        $remainder = join q{/}, @dirs;

    }

    $dir_show .= qq{/$efs};
    $dir_show .= qq{/$remainder} if $remainder;
    $dir_show =~ s{/+}{/}gmsx;

    return $dir_show;

}

sub display_translate {

    my $class = shift;

    my ( $key, $value ) = @_;

    if ( $key eq q{value} && $value =~ m/\A \d{10} \z/msx ) {
        return time2str(q{%Y-%m-%d %H:%M:%S (GMT)}, $value, q{GMT});
    } else {
        return $value;
    }

}

sub time2dbstr {

    my $class = shift;
    my (%args) = @_;

    my $dbvendor = $class->get_db_vendor;

    my %format = (
        mysql  => {
            date => q{%Y-%m-%d},
            time => q{%H:%M:%S},
        },
        oracle => {
            date => q{%d-%b-%y},
            time => q{%H.%M.%S},
        },
    );

    my $format = $format{$dbvendor}->{date};
    if ( $args{type} eq q{time} ) {
        $format .= q{ } . $format{$dbvendor}->{time};
    }

    my $datetime = time2str( $format, $args{time}, q{GMT} );

    return $datetime;

}

sub get_db_vendor {
    my $class = shift;
    my $connect = $class->get_db_connect;
    my ($vendor) = (split m{:}msx, $connect)[1];
    return lc($vendor);
}

sub get_id_column_attrs {
    my $class = shift;
    my $dbvendor = $class->get_db_vendor;
    if ( $dbvendor eq q{oracle} ) {
        return (
            auto_nextval => 1,
        );
    }
    else {
        return;
    }
}

sub set_db_password {

    my $class = shift;
    my (%args) = @_;

    my $dbvendor = $class->get_db_vendor;
    my $dbh = $class->get_schema->storage->dbh;

    given ( $dbvendor ) {
        when ( q{oracle} ) {
            $dbh->do( qq{alter user $args{user} identified by $args{password}} );
        }
        when ( q{mysql} ) {
            $dbh->do( qq{set password = password('$args{password}')} );
        }
        default {
            die qq{Unsupported vendor type '$dbvendor'\n};
        }
    }

    return;

}

sub get_db_prefix {
    if ( -f q{/etc/efs/server.conf} ) {
        my $config = EFS::Config::Server->new;
        return $config->get_value( q{database}, q{prefix} );
    }
    return;
}

{

    my %quote_char = (
        oracle => q{"},
        mysql   => q{`},
        sqlite  => q{"},
    );

    sub get_db_quote_char {
        my $class = shift;
        my $dbvendor = $class->get_db_vendor;
        return $quote_char{$dbvendor};
    }

}

{

    my %schema_of         = ();
    my %schema_lastaccess = ();

    sub get_schema {

        my $class = q{EFS::DBI};

        if ( $schema_of{$PID} ) {

            my $dbh;

            try { $dbh = $schema_of{$PID}->storage->dbh; };

            if ($dbh) {

                # XXX: This nonsense is Oracle-specific, and a workaround for
                # unstable connections
                if ( $dbh->ping ) {
                    $schema_lastaccess{$PID} = time;
                    return $schema_of{$PID};
                }

                my $elapsed = time - $schema_lastaccess{$PID};
                warn(
                    q{Re-connecting to database. },
                    qq{Ping failed, handle last accessed $elapsed secs ago.\n},
                    qq{DBI Error: $DBI::errstr\n},
                );

            }

            delete $schema_of{$PID};
            delete $schema_lastaccess{$PID};

        }

        my $dbvendor = $class->get_db_vendor;
        my $connect = $class->get_db_connect;
        my $prefix = $class->get_db_prefix || q{efs};
        my $db_user = $class->get_db_user;
        my @passwords = $class->get_db_passwords($db_user);

        my @errors = ();

PASSWORD:
        foreach my $password (@passwords) {

            my $schema = $class->connect(
                $connect, $db_user, $password,
                {
                    AutoCommit => 1,
                    RaiseError => 1,
                },
                {
                    on_connect_call => q{datetime_setup},
                    quote_char => $class->get_db_quote_char,
                    name_sep => q{.},
                },
            );

            my $dbh;

            my $dbh_error_occured;

            try { $dbh = $schema->storage->dbh; }
            catch {
                push @errors, $DBI::errstr, $_;
                $dbh_error_occured = 1;
            };
            next PASSWORD if $dbh_error_occured;

            if ( $dbvendor eq q{sqlite} ) {
                my $dbname =
                    q{/efs/dist/efs/config-database/current/common/efs.db};
                $dbh->do(qq{ATTACH DATABASE "$dbname" AS $prefix});
            }

            if ( ref $schema && $schema->isa(q{DBIx::Class::Schema}) ) {
                $schema_of{$PID} = $schema;
                last;
            }

        }

        if ( not $schema_of{$PID} ) {
            die(
                qq{Unable to connect to EFS database:\n},
                @errors,
            );
        }

        $schema_lastaccess{$PID} = time;

        return $schema_of{$PID};

    }

    #
    # After each and every fork, this attribute MUST be set on any
    # inherited DBD::Oracle handles, otherwise when the subprocess
    # exits, they will be destroyed.  When they are destroyed, errors
    # will result, since the handles are invalid, and you can't
    # connect to the database to logout cleanly.
    #
    # One of these days, this will be addressed internally in
    # DBD::Oracle, but until then we MUST own every fork call.
    #

    sub inactivate_handles {

        my $class = shift;

        # XXX: This may not be necessary anymore, since it appears dbix
        # doesn't keep the connections open.

        foreach my $pid ( keys %schema_of ) {
            next if $pid == $PID;
            if ( $schema_of{$pid}->storage->connected ) {
                $schema_of{$pid}->storage->dbh->{InactiveDestroy} = 1;
            }
        }

        return;
    }

}

{

    my $db_user = q{};

    sub set_db_user {
        my $class = shift;
        $db_user = shift;
        return;
    }

    sub get_db_user {

        my $class = shift;

        if ( -f q{/etc/efs/server.conf} && $EUID == 0 ) {
            if ( not $db_user ) {
                my $config = EFS::Config::Server->new;
                $db_user = $config->get_value( q{database}, q{rwuser} );
            }
        }
        else {
            $db_user = q{sqlite};
        }

        return $db_user;

    }

}

{

    my $db_connect = q{};

    sub set_db_connect {
        my $class = shift;
        $db_connect = shift;
        return;
    }

    sub get_db_connect {

        my $class = shift;

        if ( -f q{/etc/efs/server.conf} && $EUID == 0 ) {
            if ( not $db_connect ) {
                my $config = EFS::Config::Server->new;
                $db_connect = $config->get_value( q{database}, q{connect} );
            }
        }
        else {
            $db_connect = q{DBI:SQLite:dbname=:memory:};
        }

        return $db_connect;

    }

}

sub get_db_passwords {

    my $class    = shift;
    my $username = shift;

    if ( $username eq q{sqlite} ) {
        return q{sqlite};
    }

    my $password_dir = q{/var/efs/etc/passwd};

    my $password_file = qq{$password_dir/$username};

    my @passwords = ();

    my $io = IO::File->new($password_file)
        || die qq{Unable to open $password_file: $ERRNO\n};

    while ( defined ( $_ = $io->getline ) ) {
        s/\A \h+//msx;
        s/\h+ \z//msx;
        chomp;
        push @passwords, $_;
    }

    if ( not @passwords ) {
        die qq{Empty password file: $password_file\n};
    }

    return @passwords;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
