package EFS::DBI::Result::ReleaseAlias;

use strict;
use warnings;
use English;

use base q{EFS::DBI::Result};
use base q{EFS::DistableObject};

__PACKAGE__->load_components( qw/+EFS::DBI::Prefix Core/ );

__PACKAGE__->table( q{releasealias} );

__PACKAGE__->add_columns( qw( project name ) );
__PACKAGE__->add_columns(
    id => {
        __PACKAGE__->get_id_column_attrs,
    },
    target => {
        accessor => q{_target},
    },
);
__PACKAGE__->set_primary_key( q{id} );
__PACKAGE__->sequence( q{releasealias_seq} );

__PACKAGE__->belongs_to( project => q{EFS::DBI::Result::Project} );

__PACKAGE__->has_many( cellstates => q{EFS::DBI::Result::ReleaseAliasCellState} );
__PACKAGE__->has_many( attrs => q{EFS::DBI::Result::ReleaseAliasAttrs} );

sub metaproj {
    return shift->project->metaproj;
}

sub cell {
    return shift->project->cell;
}

sub release {
    my $self = shift;
    my $resultset = $self->get_resultset( q{Release} );
    my ($release) = $resultset->search({
        project => $self->project->id,
        name    => $self->_target,
    });
    return $release;
}

sub releasealias {
    my $self = shift;
    my $resultset = $self->get_resultset( q{ReleaseAlias} );
    my ($releasealias) = $resultset->search({
        project => $self->project->id,
        name    => $self->_target,
    });
    return $releasealias;
}

sub installs {
    return shift->target_release->installs;
}

sub target {
    my $self = shift;
    return $self->release || $self->releasealias;
}

sub releasealiases {
    my $self = shift;
    my $resultset = $self->get_resultset( q{ReleaseAlias} );
    my @releasealiases = $resultset->search({
        project => $self->project->id,
        target  => $self->name,
        @_,
    });
    return @releasealiases;
}

sub container {
    return shift->target;
}

sub all_releasealiases {

    my $self = shift;

    my @results = ();

    my $resultset = $self->get_resultset( q{ReleaseAlias} );

    my @releasealiases = $resultset->search({ project => $self->project->id });

    foreach my $releasealias ( @releasealiases ) {

        if ( $releasealias->name eq $self->name ) {
            next;
        }

        if ( not $releasealias->resolves_to_name( $self->name ) ) {
            next;
        }

        push @results, $releasealias;

    }

    return @results;

}

#
# Walk the chain of rl->rl->rl->release to get to the release at the end
#
sub target_release {

    my $object = shift;

    my $release_rs      = $object->get_resultset( q{Release} );
    my $releasealias_rs = $object->get_resultset( q{ReleaseAlias} );

    while ( $object->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {

        my $previous = $object;

        my $previous_mpr_target = $previous->target->object_id;

        ($object) = $release_rs->search({
            project => $previous->project->id,
            name    => $previous->target->name,
        });

        if ( not $object ) {
            ($object) = $releasealias_rs->search({
                project => $previous->project->id,
                name    => $previous->target->name,
            });
        }

        if ( not $object ) {
            die(
                q{Unable to retrieve release(link) object },
                qq{for $previous_mpr_target\n}
            );
        }

    }

    return $object;

}

# This is a boolean, true if the releasealias eventually resolves to
# the given name

sub resolves_to_name {

    my $object = shift;
    my $name   = shift;

    my $release_rs      = $object->get_resultset( q{Release} );
    my $releasealias_rs = $object->get_resultset( q{ReleaseAlias} );

    while ( $object->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {

        my $previous = $object;

        my $previous_mpr_target = $previous->target->object_id;

        if ( $object->target->name eq $name ) {
            return 1;
        }

        ($object) = $release_rs->search({
            project => $previous->project->id,
            name    => $previous->target->name,
        });

        if ( $object ) {
            last;
        }

        ($object) = $releasealias_rs->search({
            project => $previous->project->id,
            name    => $previous->target->name,
        });

        if ( not $object ) {
            die(
                q{Unable to retrieve release(link) object },
                qq{for $previous_mpr_target\n}
            );
        }

    }

    if ( $object->name eq $name ) {
        return 1;
    } else {
        return 0;
    }

}

sub stage {
    return shift->target_release->stage;
}

sub dependents {
    my $self = shift;
    my $resultset = $self->get_resultset( q{DependencyReleaseAlias} );
    my @dependents = $resultset->search({
        releasealias => $self->id,
        @_,
    });
    return @dependents;
}

sub releasealias_dependents {

    my $self = shift;

    my @releasealiases = (
        $self,
        $self->all_releasealiases,
    );

    my @dependents = ();

    my $resultset = $self->get_resultset( q{DependencyReleaseAlias} );

    foreach my $releasealias ( @releasealiases ) {
        push( @dependents,
              $resultset->search({
                  releasealias   => $releasealias->id,
                  @_,
              })
          );
    }

    return @dependents;

}

sub all_dependents {
    my $self = shift;
    return (
        $self->dependents(@_),
        $self->releasealias_dependents(@_),
    );

}

sub get_path {
    my $self = shift;
    my $pathtype = shift || q{dist};
    return $self->project->get_path($pathtype) . q{/} . $self->name;
}

sub assert_not_disabled {

    my $self = shift;

    if ( my $disabled = $self->is_disabled ) {
        die qq{Release '$self' is disabled: $disabled\n};
    }

    return 1;

}

sub object_id {
    my $self = shift;
    my $separator = shift || q{/};
    return join( $separator,
                 $self->metaproj->name,
                 $self->project->name,
                 $self->name );
}

sub target_for_cell {

    my $self = shift;
    my $cell = shift;

    my ($cellstate) = $self->cellstates({ cell => $cell->id }) or return;

    return $cellstate->target;
}

sub get_diststate {
    return EFS::Diststate::ReleaseAlias->new({ releasealias => shift });
}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
