package EFS::DBI::ResultSet::Cell;

use strict;
use warnings;
use English;

use base q{EFS::DBI::ResultSet};

use Params::Validate qw( :all );
use Readonly;

Readonly my %IS_VALID_CELLTYPE => map { ( $_ => 1 ) } qw{ dev test prod };
Readonly my %IS_VALID_FSTYPE   => map { ( $_ => 1 ) } qw{ afs nfs };

sub validate_data {

    my $class = shift;

    my %data = @_;

    if ( exists $data{celltype} and defined $data{celltype}
            and not exists $IS_VALID_CELLTYPE{ $data{celltype} } ) {
        die(
            qq{Invalid celltype '$data{celltype}'\n},
            q{Must be one of: },
            join( q{, }, keys %IS_VALID_CELLTYPE ),
            qq{\n},
        );
    }

    if ( exists $data{fstype} and defined $data{fstype}
            and not exists $IS_VALID_FSTYPE{ $data{fstype} } ) {
        die(
            qq{Invalid fstype '$data{fstype}'\n},
            q{Must be one of: },
            join( q{, }, keys %IS_VALID_FSTYPE ),
            qq{\n},
        );
    }

    return 1;
}

sub retrieve_bynames_assert {

    my $self = shift;

    my (%args) = validate( @_, {
        name => { type => SCALAR },
    });

    $args{name} = $self->expand_cell_name( %args );

    my ($cell) = $self->search({ name => $args{name} });

    if ( not $cell ) {
        die qq{No such cell '$args{name}'\n};
    }

    return $cell;

}

sub expand_cell_name {

    my $self = shift;
    my %args = @_;

    my $parameter_rs = $self->get_resultset( q{Parameter} );

    my $root = $parameter_rs->configured( q{root_domain} ) ||
        die qq{No root_domain parameter has been defined.\n};

    if ( $args{name} =~ m/$root \z/msx ) {
        return $args{name};
    }

    if ( exists $args{name} &&
             exists $args{location} &&
                 exists $args{campus} &&
                     exists $args{region} ) {

        if ( $args{name} !~ m/\./msx ) {
            $args{name} = join( q{.}, @args{qw( name location campus region )}, $root );
        }

    } else {

        my @cells = $self->search({ name => { -like => qq{$args{name}%} } });

        if ( not @cells) {
            die qq{No cell name matching '$args{name}*'\n};
        }

        if ( @cells > 1 ) {
            die(
                qq{More than one cell matches '$args{name}*':\n},
                join( qq{\n\t}, map { $_->name } @cells ), qq{\n},
            );
        }

        $args{name} = $cells[0]->name;

    }

    return $args{name};

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
