package EFS::DBI::ResultSet::Release;

use strict;
use warnings;
use English;

use base q{EFS::DBI::ResultSet};

use Params::Validate qw( :all );
use Readonly;

Readonly my %IS_VALID_STAGE => map { ( $_ => 1 ) }
    qw{ dev qa rejected prod deprecated };

sub validate_data {

    my $class = shift;
    my %data  = @_;

    if ( exists $data{release} ) {
        $data{release} =~ m/\A [-\w\.]+ \z/msx || die(
            qq{Invalid release name '$data{release}'\n},
            q{Must consist solely of alphanumeric characters, },
            qq{underscores, dashes and/or dots\n}
        );
        EFS::DBI::ResultSet::ReleaseAlias->assert_not_reserved($data{release});
    }

    if ( exists $data{stage}
            and not exists $IS_VALID_STAGE{ $data{stage} } ) {
        die(
            qq{Invalid stage name '$data{stage}'\n},
            q{Must be one of: },
            join( q{, }, keys %IS_VALID_STAGE ),
            qq{.\n},
        );
    }

    return 1;

}

sub retrieve_release_or_alias_assert {

    my $self = shift;

    my (%args) = validate( @_, {
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $object = $self->retrieve_release_or_alias( %args );

    if ( not $object ) {
        die(
            q{No such release(alias) },
            qq{$args{metaproj}/$args{project}/$args{name}.\n}
        );
    }

    return $object;

}

sub retrieve_release_or_alias {

    my $self = shift;

    my (%args) = validate( @_, {
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        name     => { type => SCALAR },
    });

    my $project_rs = EFS::DBI->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    if ( not $project ) {
        return;
    }

    my $object = q{};

    foreach my $try_class ( qw( Release ReleaseAlias ) ) {
        my $resultset = EFS::DBI->get_resultset( $try_class );
        ($object) = $resultset->search({
            project => $project->id,
            name    => $args{name},
        });
        last if $object;
    }

    return $object;

}

sub retrieve_bynames_assert {

    my $self = shift;

    my (%args) = validate( @_, {
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        name      => { type => SCALAR },
        disabled  => { type => BOOLEAN, optional => 1 },
        destroyed => { type => BOOLEAN, optional => 1 },
    });

    my $project_rs = EFS::DBI->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
        disabled => $args{disabled},
    );

    my ($release) = $self->search({
        project => $project->id,
        name    => $args{name},
    });

    if ( not $release ) {
        die qq{No such release '$project/$args{name}'\n};
    }

    if ( not $args{disabled} ) {
        $release->assert_not_disabled;
    }

    if ( not $args{destroyed} ) {
        $release->assert_not_destroyed;
    }

    return $release;

}

sub retrieve_bynames {

    my $self = shift;

    my (%args) = validate( @_, {
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        name      => { type => SCALAR },
        disabled  => { type => BOOLEAN, optional => 1 },
        destroyed => { type => BOOLEAN, optional => 1 },
    });

    my $project_rs = EFS::DBI->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames(
        metaproj => $args{metaproj},
        name     => $args{project},
        disabled => $args{disabled},
    );

    if ( not $project ) {
        return;
    }

    my ($release) = $self->search({
        project => $project->id,
        name    => $args{name},
    });

    return $release;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
