package EFS::DependencyTree;

use strict;
use warnings;
use English;

use Moose;
use Moose::Util::TypeConstraints qw(enum); # don't export type() to my namespace
use Scalar::Util q{refaddr};
use Params::Validate qw( :all );
use Clone qw(clone);
use List::MoreUtils qw( uniq );
use Readonly;
use Term::ANSIColor qw(colored);
use Try::Tiny;
$Term::ANSIColor::EACHLINE = qq{\n};

has q{type} => (
    is      => q{ro},
    isa     => enum( [ qw(runtime build test) ] ),
    default => q{runtime},
);

has q{kind} => (
    is      => q{ro},
    isa     => enum( [ qw(dependencies dependents) ] ),
    default => q{dependencies},
);

has q{seeds} => (
    is         => q{ro},
    isa        => q{ArrayRef},
    required   => 1,
    auto_deref => 1,
);


has [qw( depth_of _dependencies _dependents _seen _dependency_roots _dependent_roots _diststate )] => (
    is      => q{ro},
    isa     => q{HashRef},
    default => sub { {} },
);

has [ qw( maxdepth_of dist_failed_of dist_succeeded_of ) ] => (
    is => q{rw},
);

has [qw( objects_at_depth_of )] => (
    is      => q{rw},
    isa     => q{HashRef},
    default => sub { {} },
);

Readonly my $SCREEN_WIDTH => 1024;

{


    sub BUILD {

        my $self  = shift;

      SEED: foreach my $seed ( $self->seeds ) {

            if ( not ( $seed->isa(q{EFS::DBI::Result::ReleaseAlias}) or
                         $seed->isa(q{EFS::DBI::Result::Release}) ) ) {
                die(
                    q{Invalid argument: root must be either a },
                    qq{release or releasealias object,\n},
                    q{not a }, ref $seed, qq{object\n},
                );
            }

            if ( $self->_seen->{ $seed->object_id } ) {
                next SEED;
            }

            if ( $self->kind eq q{dependencies} ) {
                $self->expand_dependencies( object => $seed );
            } else {
                $self->expand_dependents( object => $seed );
            }

        }

        $self->check_circular;

        $self->find_roots;

        $self->get_dependency_depths;

        return;
    }

    sub has_dependency {
        my $self        = shift;
        my $object_id   = shift;
        return $self->_seen->{$object_id};
    }

    sub has_dependent {
        my $self        = shift;
        my $object_id   = shift;
        return $self->_seen->{$object_id};
    }

    sub expand_dependencies {

        my $self = shift;

        my (%args) = validate( @_, {
            object => { type => OBJECT },
        });

        my $object    = $args{object};
        my $object_id = $object->object_id;

        $self->_seen->{ $object_id } = $object;

        if ( $object->isa(q{EFS::DBI::Result::ReleaseAlias}) ) {

            my $target    = $object->target;
            my $target_id = $target->object_id;

            $self->_dependencies->{ $object_id }->{ $target_id }++;

            $self->_dependents->{ $target_id }->{ $object_id }++;

            if ( not $self->_seen->{ $target_id } ) {
                $self->expand_dependencies( object => $target );
            }

        } else {

            my @dependencies = $object->dependencies(
                type => $self->type,
            );

            foreach my $dependency ( @dependencies ) {

                my $target    = $dependency->target;
                my $target_id = $target->object_id;

                $self->_dependencies->{ $object_id }->{ $target_id }++;

                $self->_dependents->{ $target_id }->{ $object_id }++;

                if ( not $self->_seen->{ $target_id } ) {
                    $self->expand_dependencies( object => $target );
                }

            }

        }

        return;
    }

    sub expand_dependents {

        my $self = shift;

        my (%args) = validate( @_, {
            object => { type => OBJECT },
        });

        my $object    = $args{object};
        my $object_id = $object->object_id;

        $self->_seen->{ $object_id } = $object;

        my @releasealiases = $object->releasealiases;

        foreach my $releasealias ( @releasealiases ) {

            my $releasealias_id = $releasealias->object_id;

            $self->_dependencies->{ $releasealias_id }->{ $object_id }++;

            $self->_dependents->{ $object_id }->{ $releasealias_id }++;

            if ( not $self->_seen->{ $releasealias_id } ) {
                $self->expand_dependents( object => $releasealias );
            }

        }

        # NOTE: @dependencies objects are *relationships*
        my @dependencies = $object->dependents(
            type => $self->type,
        );

        foreach my $dependency ( @dependencies ) {

            my $dependent    = $dependency->dependent_release;
            my $dependent_id = $dependent->object_id;

            $self->_dependencies->{ $dependent_id }->{ $object_id }++;

            $self->_dependents->{ $object_id }->{ $dependent_id }++;

            if ( not $self->_seen->{ $dependent_id } ) {
                $self->expand_dependents( object => $dependent );
            }

        }

        return;

    }

    sub find_conflicts {

        my $self = shift;

        my (%args) = validate( @_, {
            interested => { type     => HASHREF,
                            optional => 1 },
        });

        my $interested  = $args{interested};

        my %releases_of = ();
        my %conflicts   = ();

        foreach my $object_id ( keys %{ $self->_seen } ) {

            my ($metaproj,$project) = split( m{/}msx, $object_id );

            if ( $interested and
                     not $interested->{ $metaproj } and
                         not $interested->{ qq{$metaproj/$project} } ) {
                next;
            }

            my $object = $self->_seen->{ $object_id };

            my $target_id = $object_id;

            if ( $object->isa(q{EFS::DBI::Result::ReleaseAlias}) ) {
                $target_id = $object->target_release->object_id;
            }

            $releases_of{ qq{$metaproj/$project} }->{ $target_id }++;

        }

        foreach my $mp_name ( keys %releases_of ) {
            if ( scalar keys %{ $releases_of{ $mp_name } } > 1 ) {
                foreach my $mpr_name ( keys %{ $releases_of{ $mp_name } } ) {
                    $conflicts{ $mpr_name } = q{conflict};
                }
            }
        }

        return %conflicts;

    }

    sub check_circular {

        my $self = shift;

        EFS::DBI::Dependency->check_circular(
            dependencies => clone($self->_dependencies),
            dependents   => clone($self->_dependents),
        );

        return 1;

    }

    sub find_roots {

        my $self = shift;

        foreach my $dependency_mpr ( keys %{ $self->_dependencies } ) {
            if ( not exists $self->_dependents->{ $dependency_mpr } ) {
                $self->_dependency_roots->{ $dependency_mpr }++;
            }
        }

        foreach my $dependent_mpr ( keys %{ $self->_dependents } ) {
            if ( not exists $self->_dependencies->{ $dependent_mpr } ) {
                $self->_dependent_roots->{ $dependent_mpr }++;
            }
        }

        foreach my $seed ( $self->seeds ) {

            my $seed_id = $seed->object_id;

            if ( not exists $self->_dependents->{ $seed_id } ) {
                $self->_dependency_roots->{ $seed_id }++;
            }

            if ( not exists $self->_dependencies->{ $seed_id } ) {
                $self->_dependent_roots->{ $seed_id }++;
            }

        }

        return 1;

    }

    sub get_seen {
        my $self = shift;
        return @{[ sort { lc($a) cmp lc($b) } keys %{ $self->_seen } ]};
    }

    sub get_tree_text {

        my $self = shift;

        my (%args) = validate( @_, {
            highlight => { type     => HASHREF,
                           default  => {},
                           optional => 1 },
            maxdepth  => { type     => SCALAR,
                           default  => 0,
                           optional => 1 },
            cells     => { type     => ARRAYREF,
                           optional => 1 },
        });

        $self->get_diststates;

        $args{cells} ||= [ EFS::DBI->handle_cell_arguments(
            celltype        => q{prod},
            include_parents => 1,
            default_scope   => q{global}, # XXX: Not sure about this
        ) ];

        my $table = Text::FormatTable->new( q{l  r  l  l} );
        $table->head( q{Stage}, q{Depth}, q{Diststate},
                      join(q{ },
                           ucfirst($self->type),
                           ucfirst($self->kind)) );
        $table->rule( q{=} );

        my @roots = ();

        if ( $self->kind eq q{dependencies} ) {
            @roots =
                sort { lc($a) cmp lc($b) }
                    keys %{ $self->_dependency_roots };
        } else {
            @roots =
                sort { lc($a) cmp lc($b) }
                    keys %{ $self->_dependent_roots };
        }

        my $seen = {};

        foreach my $root ( @roots ) {

            $self->get_tree_text_nextrow(
                object    => $self->_seen->{$root},
                table     => $table,
                highlight => $args{highlight},
                depth     => 1,
                maxdepth  => $args{maxdepth},
                seen      => $seen,
                cells     => $args{cells},
            );

        }

        my $text = $table->render($SCREEN_WIDTH);

        $text =~ s{__}{  }gmsx;

        return $text;

    }

    sub get_tree_text_nextrow {

        my $self  = shift;

        my (%args) = validate( @_, {
            object    => { type => OBJECT },
            table     => { isa  => q{Text::FormatTable} },
            highlight => { type => HASHREF },
            depth     => { type => SCALAR },
            maxdepth  => { type => SCALAR },
            seen      => { type => HASHREF },
            cells     => { type => ARRAYREF },
        });

        my $object    = $args{object};
        my $table     = $args{table};
        my $highlight = $args{highlight};
        my $seen      = $args{seen};
        my $depth     = $args{depth};

        my $indent    = q{__} x $depth;

        my $object_id    = $object->object_id;
        my $object_stage = $object->stage;

        my $shorthand = $self->get_diststate_shorthand(
            object_id => $object_id,
            cells     => $args{cells},
        );

        if ( $seen->{$object_id} ||
                 ( $self->kind eq q{dependencies} &&
                       $depth != $self->depth_of->{$object_id} ) ) {
            my $entry = $indent . $object_id;
            my $color;
            if ( $highlight->{$object_id} ) {
                $entry .= qq{ ($highlight->{$object_id}, repeat)};
                $color = q{red};
            } else {
                $entry .= q{ (repeat)};
                $color = q{yellow};
            }
            $table->row( $object_stage, $depth,
                         $shorthand, colored( $entry, $color ) );
            return 1;
        }

        if ( $self->kind ne q{dependencies} ||
                 $depth == $self->depth_of->{$object_id} ) {
            $seen->{$object_id} = $object;
        }

        if ( $highlight->{$object_id} ) {
            my $entry = $indent . $object_id . qq{ ($highlight->{$object_id})};
            $table->row( $object_stage, $depth,
                         $shorthand, colored($entry, q{red}) );
        } else {
            $table->row( $object_stage, $depth,
                         $shorthand, $indent . $object_id );
        }

        if ( $args{maxdepth} > 0 && $depth == $args{maxdepth} ) {
            return 1;
        }

        my @children = ();

        if ( $self->kind eq q{dependencies} ) {
            @children =
                sort { lc($a) cmp lc($b) }
                    keys %{ $self->_dependencies->{$object_id} };
        } else {
            @children =
                sort { lc($a) cmp lc($b) }
                    keys %{ $self->_dependents->{$object_id} };
        }

        foreach my $child_id ( @children ) {
            $self->get_tree_text_nextrow(
                object    => $self->_seen->{$child_id},
                table     => $table,
                highlight => $highlight,
                depth     => $depth + 1,
                maxdepth  => $args{maxdepth},
                seen      => $seen,
                cells     => $args{cells},
            );
        }

        return;
    }

    # XXX: This should probably be in Diststate.pm
    sub get_diststate_shorthand {

        my $self = shift;
        my %args = @_;

        my $object_id = $args{object_id};
        my $object    = $self->_seen->{$object_id};

        my $diststate = $self->_diststate->{$object_id};

        if ( $object->isa(q{EFS::DBI::Result::Release}) and
                 not $diststate->get_checktime ) {
            return q{nocp};
        }

        my @cells = ();

        if ( $object->isa(q{EFS::DBI::Result::Release}) ) {
            @cells = $diststate->cells_needing_update( cells => $args{cells} );
        } else {
            @cells = $diststate->cells_never_updated( cells => $args{cells} );
        }

        my %typecount = ();

        foreach my $cell ( @cells ) {
            $typecount{ $cell->celltype }++;
        }

        my @shorthand = ();

        foreach my $celltype ( sort keys %typecount ) {
            push( @shorthand, $typecount{$celltype} . substr($celltype,0,1) );
        }

        return join( q{}, @shorthand );

    }

    sub get_dependency_depths {

        my $self  = shift;

        my @roots =
            sort { lc($a) cmp lc($b) } keys %{ $self->_dependency_roots };

        $self->objects_at_depth_of->{ 1 } = { map { $_ => 1 } @roots };

        $self->maxdepth_of( 1 ); # Initialize

        foreach my $root ( @roots ) {

            $self->depth_of->{$root} = 1;

            $self->get_dependency_depths_next_layer(
                object_id => $root,
                depth     => 2,
            );

        }

        foreach my $object_id ( keys %{ $self->depth_of } ) {
            my $depth = $self->depth_of->{$object_id};
            $self->objects_at_depth_of->{ $depth }->{ $object_id }++;
        }

        return;
    }

    sub get_dependency_depths_next_layer {

        my $self  = shift;

        my (%args) = validate( @_, {
            object_id => { type => SCALAR },
            depth     => { type => SCALAR },
        });

        my $depth     = $args{depth};
        my $object_id = $args{object_id};

        my @children = sort keys %{ $self->_dependencies->{$object_id} };

        if ( @children ) {
            my $curr_maxdepth_of = $self->maxdepth_of;
            if ( !$curr_maxdepth_of || $curr_maxdepth_of < $depth ) {
                $self->maxdepth_of( $depth );
            }
        }

        foreach my $child ( @children ) {

            $self->depth_of->{$child} ||= $depth;

            if ( $self->depth_of->{$child} < $depth ) {
                $self->depth_of->{$child} = $depth;
            }

            $self->get_dependency_depths_next_layer(
                object_id => $child,
                depth     => $depth + 1,
            );

        }

        return;
    }

    sub get_diststates {

        my $self = shift;

        foreach my $object_id ( keys %{ $self->_seen } ) {
            my $object = $self->_seen->{$object_id};
            $self->_diststate->{$object_id} = $object->get_diststate;
        }

        return;
    }

    sub authorize {

        my $self = shift;
        my %args = @_;

        my @objects = ();

        foreach my $object_id ( keys %{ $self->_seen } ) {

            my $object    = $self->_seen->{$object_id};
            my $diststate = $self->_diststate->{$object_id};

            my @needs_update = ();

            if ( $object->isa(q{EFS::DBI::Result::Release}) ) {
                @needs_update =
                    $diststate->needs_update( cells => $args{cells} );
            } else {
                @needs_update =
                    $diststate->cells_never_updated( cells => $args{cells} );
            }

            if ( @needs_update ) {
                push( @objects, $object );
            }

        }

        my $errors = q{};

        foreach my $object ( @objects ) {

            try {
                EFS::DBI->authorize(
                    noefsops => $args{noefsops},
                    authuser => $args{authuser},
                    object   => $object,
                    action   => q{disttree},
                    type     => $object->class2type,
                    history  => {
                        metaproj => $object->metaproj->name,
                    },
                );
            }
            catch {
                $errors .= $_;
            };

        }

        if ( $errors ) {
            die(
                qq{Access denied.\n},
                qq{Insufficient access to disttree dependencies:\n},
                $errors,
            );
        }

        return 1;

    }

    sub dist {

        my $self  = shift;

        my %args = @_;

        my @cells = EFS::DBI->handle_cell_arguments(
            %args,
            include_parents     => 1,
        );

        $args{cells} = [ map { $_->name } @cells ];

        $self->get_diststates;

        my @prodcells = grep { $_->celltype eq q{prod} } @cells;

        if ( @prodcells ) {

            my @nonprod =
                grep { not $_->has_only_prod_stages } $self->seeds;

            if ( @nonprod ) {
                die(
                    q{Unable to distribute this dependency tree to the },
                    qq{specified prod cells.\n},
                    qq{The following release(link)s have invalid stages:\n\t},
                    join(qq{\n\t}, map { qq{$_ => } . $_->stage } @nonprod ), qq{\n},
                );
            }

        }

        $self->authorize(
            noefsops => $args{noefsops},
            authuser => $args{authuser},
            cells    => \@cells,
        );

        my $depth = $self->maxdepth_of;

        $self->dist_failed_of( {} );
        $self->dist_succeeded_of( {} );

        my @all_uptodate_ids = ();

      DEPTH: while ( $depth ) {

            my @object_ids = ();
            my @failed_ids = ();
            my @uptodate_ids = ();

            foreach my $object_id
                ( sort { lc($a) cmp lc($b) }
                      keys %{ $self->objects_at_depth_of->{$depth} } ) {

                my $diststate = $self->_diststate->{$object_id};
                my $object    = $self->_seen->{$object_id};

                if ( $self->dist_failed_of->{$object_id} ) {
                    push( @failed_ids, $object_id );
                    next;
                }

                my @needs_update = ();

                if ( $object->isa(q{EFS::DBI::Result::Release}) ) {
                    @needs_update =
                        $diststate->needs_update( cells => \@cells );
                } else {
                    @needs_update =
                        $diststate->cells_never_updated( cells => \@cells );
                }

                if ( @needs_update ) {
                    push( @object_ids, $object_id );
                } else {
                    push( @uptodate_ids, $object_id );
                    push( @all_uptodate_ids, $object_id );
                }

            }

            my $failed_count   = $#failed_ids + 1;
            my $uptodate_count = $#uptodate_ids + 1;
            my $object_count   = $#object_ids + 1;

            if ( $failed_count ) {
                my $mprs = $failed_count > 1 ? q{MPRs} : q{MPR};
                STDOUT->print(
                    qq{Skipping $failed_count $mprs at depth $depth, },
                    qq{due to dependent failures\n},
                );
            }

            if ( $uptodate_count ) {
                my $mprs = $uptodate_count > 1 ? q{MPRs} : q{MPR};
                my $are  = $uptodate_count > 1 ? q{are} : q{is};
                STDOUT->print(
                    qq{Skipping $uptodate_count $mprs at depth $depth, },
                    qq{which $are up to date\n},
                );
            }

            if ( $object_count ) {
                my $mprs = $object_count > 1 ? q{MPRs} : q{MPR};
                STDOUT->print(
                    qq{Distributing $object_count $mprs at depth $depth:\n\t},
                    join(qq{\n\t},@object_ids), qq{\n},
                );
            } else {
                STDOUT->print(
                    qq{Nothing left to distribute at depth $depth\n},
                );
                $depth--;
                next DEPTH;
            }

            my $results = EFS::RPC->execute(
                class          => $self,
                method         => q{dist_objects},
                args           => [
                    history    => $args{history},
                    object_ids => \@object_ids,
                    distargs   => \%args,
                ],
                force_remote   => not $PERLDB,
            );

            foreach my $succeeded_id ( @{ $results->{succeeded} } ) {
                $self->dist_succeeded_of->{$succeeded_id}++;
            }

            foreach my $failed_id ( @{ $results->{failed} } ) {
                $self->set_dependents_failed( $failed_id );
            }

            $depth--;

        }

        my @dist_succeeded =
            sort { lc($a) cmp lc($b) } keys %{ $self->dist_succeeded_of };
        my @dist_failed =
            sort { lc($a) cmp lc($b) } keys %{ $self->dist_failed_of };
        my @dist_uptodate =
            sort { lc($a) cmp lc($b) } @all_uptodate_ids;

        if ( @dist_uptodate ) {
            STDOUT->print(
                qq{Skipped the following up to date MPRs:\n\t},
                join( qq{\n\t}, @dist_uptodate ), qq{\n},
            );
        }

        if ( @dist_succeeded ) {
            STDOUT->print(
                qq{Successfully disted the following MPRs:\n\t},
                join( qq{\n\t}, @dist_succeeded ), qq{\n},
            );
        }

        if ( @dist_failed ) {
            die(
                qq{Failed to dist the following MPRs or their dependencies:\n\t},
                join( qq{\n\t}, @dist_failed ), qq{\n},
            );
        }

        return 1;

    }

    sub dist_objects {

        my $class = shift;

        my (%args) = validate( @_, {
            history    => { type => SCALAR },
            object_ids => { type => ARRAYREF },
            distargs   => { type => HASHREF },
        });

        my $fm_objects = EFS::ForkManager->new({

            list => $args{object_ids},

            action => sub {

                my $object_id = shift;

                my ($metaproj_name,$project_name,$release_name) =
                    split m{/}msx, $object_id;

                my $release_rs = EFS::DBI->get_resultset( q{Release} );
                my $object = $release_rs->retrieve_release_or_alias_assert(
                    metaproj => $metaproj_name,
                    project  => $project_name,
                    name     => $release_name,
                );

                my $cli_class = q{EFS::CLI::Release};
                my $cli_type  = q{release};

                my %distargs = (
                    %{ $args{distargs} },
                    q{no-strict-releasealiases} => 1,
                );

                if ( $object->isa(q{EFS::DBI::Result::ReleaseAlias}) ) {
                    $distargs{_missing_cells_only}++;
                    $cli_class = q{EFS::CLI::ReleaseAlias};
                    $cli_type  = q{releasealias};
                }

                my $request =
                    qq{dist $cli_type $metaproj_name $project_name $release_name};

                return EFS::RPC->execute(
                    class  => $cli_class,
                    method => q{dist},
                    args   => [
                        metaproj => $metaproj_name,
                        project  => $project_name,
                        name     => $release_name,
                        %distargs,
                    ],
                    history      => {
                        parentid => $args{history},
                        values   => {
                            request   => $request,
                            metaproj  => $metaproj_name,
                            project   => $project_name,
                            $cli_type => $release_name,
                        },
                    },
                );

            },

        });

        $fm_objects->run;

        my $results = {
            succeeded => [ $fm_objects->get_succeeded ],
            failed    => [ $fm_objects->get_failed ],
        };

        return $results;

    }

    sub set_dependents_failed {

        my $self  = shift;

        my $object_id = shift;

        $self->dist_failed_of->{$object_id}++;

        my @dependent_ids = keys %{ $self->_dependents->{$object_id} };

        foreach my $dependent_id ( @dependent_ids ) {
            $self->set_dependents_failed( $dependent_id );
        }

        return;
    }

}

no Moose;
__PACKAGE__->meta->make_immutable;
1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
