package EFS::FileTree;

use strict;
use warnings;
use English;

use Moose;
use File::Find;
use File::LibMagic;
use File::Slurp;
use Date::Format;
use English q{-no_match_vars};
use Fcntl q{:mode};
use Params::Validate qw( :all );
use Try::Tiny;

has q{pathname} => (
    is         => q{rw},
    isa        => q{Str},
    required   => 1,
);

has q{_names} => (
    is      => q{ro},
    isa     => q{HashRef},
    default => sub { {} },
);

{

    sub BUILD {

        my $self = shift;

        my @stat_names = qw( dev ino mode nlink uid gid rdev
                             size atime mtime ctime blksize blocks );

        my $pathname = $self->pathname;
        if ( not -e $pathname ) {
            die qq{'$pathname' does not exist\n};
        }

        if ( not -d $pathname ) {
            die qq{'$pathname' is not a directory\n};
        }

        my $flm = File::LibMagic->new;

        my $want_ref = sub {

            my $name = $File::Find::name;

            my @stat_info = lstat($name) or
                die qq{Unable to stat $name: $ERRNO\n};

            my $mode = $stat_info[2];

            foreach ( @stat_names ) {
                $self->_names->{$name}->{$_} = shift @stat_info;
            }

            $self->_names->{$name}->{-d} = S_ISDIR($mode);
            $self->_names->{$name}->{-l} = S_ISLNK($mode);
            $self->_names->{$name}->{-f} = S_ISREG($mode);

            if ( $self->_names->{$name}->{-f} ) {
                $self->_names->{$name}->{magic} = $flm->describe_filename( $name );
            }

        };

        my $prep_ref = sub {
            if ( $EUID == 0 ) {
                return @_;
            }
            else {
                return grep { $_ !~ m{\. destroyed}msx } @_;
            }
        };

        find(
            {
                wanted     => $want_ref,
                preprocess => $prep_ref,
                no_chdir   => 1,
            },
            $self->pathname,
        );

        return $self;

    }

    sub get_max_ctime {

        my $self  = shift;

        my (%args) = validate( @_, {
            show => { type     => SCALAR,
                      optional => 1 },
        });

        my $max_ctime = 0;
        my %new_files = ();

        foreach my $name ( keys %{ $self->_names } ) {

            next if $self->_names->{$name}->{-l};

            if ( $self->_names->{$name}->{ctime} > $max_ctime ) {
                $max_ctime = $self->_names->{$name}->{ctime};
            }

            if ( $args{show} &&
                     $self->_names->{$name}->{ctime} > $args{show} ) {
                $new_files{$name} = $self->_names->{$name}->{ctime};
            }

        }

        if ( %new_files ) {

            my $showdate =
                time2str( q{[%Y-%m-%d %H:%M:%S] }, $args{show}, q{GMT} );

            STDOUT->print(
                qq{The last checktime for this install was $showdate\n},
                qq{The following paths have been modified since then:\n\t},
                join( qq{\n\t},
                      map { time2str( qq{[%Y-%m-%d %H:%M:%S] => $_},
                                      $new_files{$_}, q{GMT} ) }
                          sort keys %new_files ), qq{\n}
            );
        }

        if ( $max_ctime > time ) {
            warn(
                q{Maximum ctime value for }, $self->pathname,
                qq{ is in the future!!\n},
                q{Maximum ctime => },
                time2str( q{[%Y-%m-%d %H:%M:%S] }, $max_ctime ), qq{\n},
                q{Now           => },
                time2str( q{[%Y-%m-%d %H:%M:%S] }, time ), qq{\n},
            );
        }

        return $max_ctime;

    }

    sub get_dir {

        my $self  = shift;
        my $name  = shift;

        if ( not $self->_names->{$name} ) {
            die qq{'$name' not found\n};
        }

        if ( $self->_names->{$name}->{-d}
                 && !$self->_names->{$name}->{-l} ) {
            return ( $self->_names->{$name} );
        }

        return die qq{'$name' is not a directory\n};

    }

    sub get_file {

        my $self  = shift;
        my $name  = shift;

        if ( not $self->_names->{$name} ) {
            die qq{'$name' not found\n};
        }

        ##  no critic: NegativeExpressionsInUnlessAndUntilConditions
        ##  this statement is more clear as an unless clause
        if ( $self->_names->{$name}->{-l}
                 or not $self->_names->{$name}->{-f} ) {
            die qq{'$name' is not a file\n};
        }

        return $self->_names->{$name};

    }

    sub get_link {

        my $self  = shift;
        my $name  = shift;

        if ( not $self->_names->{$name}  ) {
            die qq{'$name' not found\n};
        }

        if ( not $self->_names->{$name}->{-l} ) {
            die qq{'$name' is not a link\n};
        }

        return $self->_names->{$name};

    }

    sub get_dirnames {

        my $self  = shift;

        my @names = ();

        foreach my $name ( sort keys %{ $self->_names } ) {
            if ( $self->_names->{$name}->{-d} and not
                     $self->_names->{$name}->{-l} ) {
                push @names, $name;
            }
        }

        return @names;

    }

    sub get_dircount {
        my $self  = shift;
        my $count = scalar $self->get_dirnames;
        return $count;
    }

    sub get_filenames {

        my $self  = shift;

        my @names = ();

        foreach my $name ( sort keys %{ $self->_names } ) {
            if ( $self->_names->{$name}->{-f} and not
                     $self->_names->{$name}->{-l} ) {
                push @names, $name;
            }
        }

        return @names;

    }

    sub get_filecount {
        my $self  = shift;
        my $count = scalar $self->get_filenames;
        return $count;
    }

    sub get_linknames {
        my $self  = shift;

        my @names = ();

        foreach my $name ( sort keys %{ $self->_names } ) {
            if ( $self->_names->{$name}->{-l} ) {
                push @names, $name;
            }
        }

        return @names;

    }

    sub get_linkcount {
        my $self  = shift;
        my $count = scalar $self->get_linknames;
        return $count;
    }

    sub get_spaceused {

        my $self  = shift;

        my $total;

        foreach my $name ( keys %{ $self->_names } ) {
            $total += $self->_names->{$name}->{size};
        }

        return $total;

    }

    sub get_magic {

        my $self  = shift;
        my $name  = shift;

        if ( not $self->_names->{$name} ) {
            die qq{'$name' not found\n};
        }

        if ( not $self->_names->{$name}->{-f} ) {
            die qq{'$name' is not a file\n};
        }

        return $self->_names->{$name}->{magic};

    }

    sub replace_paths {

        my $self  = shift;

        my %replace = @_;

        my @pathnames = ( $self->get_filenames, $self->get_linknames );

        my $fm_files = EFS::ForkManager->new({

            list => \@pathnames,

            action => sub {

                my $pathname = shift;

                my $contents = q{};

                if ( -l $pathname ) {
                    $contents = readlink($pathname) or
                        die qq{Unable to readlink $pathname: $ERRNO\n};
                } else {
                    $contents = read_file($pathname);
                }

                my $changed = 0;

                foreach my $old ( keys %replace ) {
                    my $new = $replace{$old};
                    if ( $contents =~ s{$old\b}{$new}gmsx ) {
                        $changed++;
                    }
                }

                if ( $changed ) {
                    if ( -l $pathname ) {
                        EFS::Util->MakeLink(
                            from => $pathname,
                            to   => $contents,
                        );
                    } else {
                        EFS::Util->WriteFileRetry(
                            pathname => $pathname,
                            contents => $contents,
                        );
                    }
                }

                return 1;

            },

        });

        $fm_files->run ||
            die(
                qq{Unable to replace paths in the following files/links:\n\t},
                join( qq{\n\t}, $fm_files->get_failed ), qq{\n},
            );

        return 1;

    }

}

no Moose;
__PACKAGE__->meta->make_immutable;
1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
