package EFS::Request::CLI;

use strict;
use warnings;
use English;
use feature q{switch};

use EFS::FQDN qw(fqdn);

use Params::Validate qw( :all );
use POSIX;
use Try::Tiny;

use Moose;
extends q{EFS::Request};

has q{efsops} => (
    is       => q{rw},
    isa      => q{Bool},
);

sub metadata_required {
    return qw( envvars cell );
}

sub operation_required {
    return qw( object_type request action );
}

sub authorize {

    my $self  = shift;

    my $metadata  = $self->metadata;
    my $operation = $self->operation;

    my $authuser = $self->authuser;

    $authuser eq q{efstestuser} and die(
        qq{The efs client is being run as efstestuser, therefore\n},
        qq{absolutely no commands will be authorized.\n},
    );

    my $object_class = $operation->{object_class};
    my $object_type  = $operation->{object_type};
    my $request      = $operation->{request};
    my $action       = $operation->{action};
    my %args         = %{ $operation->{args} };

    my $authuser_obj = EFS::Username->new({ name => $authuser });

    my $efsops = grep { $_ eq q{efsops} } $authuser_obj->groups;

    $self->efsops( $efsops ? 1 : 0 );

    if ( not $efsops or $args{noefsops} ) {
        $self->check_maintenance_mode();
    }

    my %history_columns = map { $_ => 1 } @EFS::DBI::Result::History::other_keys;
    my %history;

    foreach my $argument ( keys %args ) {
        if ( exists $history_columns{$argument} ) {
            $history{$argument} = $args{$argument};
        };
    }

    given ( $object_class ) {

        when ( m{::ChildRelease}msx ) {
            $history{release} = $args{name};
        }

        when ( m{::Dependency\z}msx ) {
            # We register the dependency stuff based on the dependent_*
            # values, but flattened to fit in to the metaproj/project/release
            # columns Also, look for the plain metaproj/project/release
            # arguments first, and then override them with the dependent_*
            # values, if found.
            my @arguments =
                qw( metaproj project release
                    dependent_metaproj dependent_project dependent_release );

            foreach my $argument (@arguments) {
                next unless exists $args{$argument};
                ( my $short_argument = $argument ) =~ s/dependent_//msx;
                $history{$short_argument} = $args{$argument};
            }
        }

        default {
            if ( exists $args{name} && $action !~ /\A(?:set|del)attr\z/msx ) {
                $history{$object_type} = $args{name};
            }
        }

    }

    my $history_rs = EFS::DBI->get_resultset(q{History});

    %history = $history_rs->validate_data(%history);

    $self->history_args(\%history);

    EFS::DBI->authorize(
        authuser => $authuser,
        action   => $action,
        type     => $object_type,
        history  => \%history,
        noefsops => $args{noefsops},
    );

    return 1;

}

sub execute {

    my $self = shift;

    if ( $self->authrequired ) {
        return $self->execute_with_authorization;
    } else {
        return $self->execute_no_authorization;
    }

}

sub execute_no_authorization {

    my $self = shift;

    my $operation = $self->operation;

    my $object_class = $operation->{object_class};
    my $action       = $operation->{action};
    my %args         = %{ $operation->{args} };

    try { $object_class->$action(%args); }
    catch {
        STDOUT->print( $_, qq{_EFS_EXIT_STATUS: 1\n} );
        exit 1;
    };

    STDOUT->print( qq{_EFS_EXIT_STATUS: 0\n} );
    exit 0;

}

sub execute_with_authorization {

    my $self = shift;

    my $logfile = $self->logfile;

    $self->authorize;

    my $metadata  = $self->metadata;
    my $operation = $self->operation;

    my $request      = $operation->{request};
    my $action       = $operation->{action};
    my $object_type  = $operation->{object_type};
    my $object_class = $operation->{object_class};
    my %args         = %{ $operation->{args} };

    if ( EFS::DBI->get_db_vendor eq q{oracle} ) {
        if ( length ($request) > 4000 ) {
            $request = substr ( $request, 0, 3999 );
        }
    }

    my %history = %{ $self->history_args };

    my $history_rs = EFS::DBI->get_resultset(q{History});

    my $history = $history_rs->create({
        command    => qq{$action $object_type},
        request    => $request,
        logfile    => $logfile->path,
        authuser   => $self->authuser,
        efshost    => fqdn,
        efsversion => $EFS::VERSION,
        pid        => $PID,
        %history,
    });

    $args{history}  = $history->id;
    $args{authuser} = $self->authuser;

    if ( $object_type =~ m/\Arelease(link)?\z/msx && $action eq q{dist} ) {
        $args{_waitfor_cell} = $metadata->{cell};
    }

    if ( $self->efsops and not $args{noefsops} ) {
        $args{authorized} = 1;
    } else {
        $args{authorized} = 0;
    }

    try { $object_class->$action(%args); }

    ## no critic: ProhibitCaptureWithoutTest
    # Actually, this perlcritic test is wrong, IM!HO.  This *is* inside a when
    # block, which should be as good as being inside a conditional.
    catch {
        given ( $_ ) {

            when ( m{_EFS_REDIRECT_CELL:\h+(\S+)}msx ) {
                my $servercell = $1;
                $history->delete;
                STDOUT->print( qq{_EFS_REDIRECT_CELL: $servercell\n} );
                exit 0;
            }

            when ( m{_EFS_REDIRECT_SERVER:\h+(\S+)}msx ) {
                my $serverhost = $1;
                $history->delete;
                STDOUT->print( qq{_EFS_REDIRECT_SERVER: $serverhost\n} );
                exit 0;
            }

            when ( $_ ne q{} ) {
                $history->status( 0 );
                $history->update;
                STDOUT->print( $_, qq{_EFS_EXIT_STATUS: 1\n} );
                exit 1;
            }

        }
    };

    $history->status( 1 );
    $history->update;
    STDOUT->print(  qq{_EFS_EXIT_STATUS: 0\n} );
    exit 0;

}

sub check_maintenance_mode {

    my $self = shift;

    my $parameter_rs = EFS::DBI->get_resultset(q{Parameter});

    if ( my $reason = $parameter_rs->configured(q{maintenance_mode}) ) {
        die(
            qq{EFS is currently in maintenance mode.\n},
            qq{Only efsops can run authorized commands.\n},
            $reason, qq{\n}
        );
    }
    return;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
