#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Errno qw(ENOMEM);
use Test::More tests => 23;
use Test::Exception;

## no critic: RequireExtendedFormatting

{

    my $simulate_failure = 0;

    ## no critic: ProhibitExplicitReturnUndef
    ## no critic: RequireLocalizedPunctuationVars
    sub myfork {
        if ( $simulate_failure ) {
            $simulate_failure = 0;
            $ERRNO = ENOMEM;
            return undef;
        } else {
            return fork();
        }
    }

    sub enable_fork_failure {
        $simulate_failure = 1;
        return;
    }

    sub disable_fork_failure {
        $simulate_failure = 0;
        return;
    }

}

BEGIN {
    *CORE::GLOBAL::fork = \&myfork;
}

use blib;
use EFS;

throws_ok {
    EFS::ForkManager->new({});
} qr{attribute.*required}ims,
    q{testing parameter validation};

throws_ok {
    EFS::ForkManager->new({
        list     => [],
        action   => sub { },
        parallel => -1,
    });
} qr{Invalid parallel value '-1'. Must be a non-negative integer}ms,
    q{testing parallel value validation};

my $fm = EFS::ForkManager->new({
    list   => [qw( a b c d )],
    action => sub { 1; },
});
isa_ok( $fm, q{EFS::ForkManager} );
ok( $fm->run(), q{testing trivial run case} );
ok( scalar $fm->get_failed == 0, q{nothing failed} );

$fm = EFS::ForkManager->new({
    list     => [qw( a b c d )],
    action   => sub { 1; },
    parallel => 2,
});
isa_ok( $fm, q{EFS::ForkManager} );
ok( $fm->run(), q{testing trivial run case, with parallel == 2} );
ok( scalar $fm->get_failed == 0, q{nothing failed} );

$fm = EFS::ForkManager->new({
    list   => [qw( a b c d )],
    action => sub {
        my ($arg) = shift;
        return ( $arg eq q{a} ? 0 : 1 );
    },
});
isa_ok( $fm, q{EFS::ForkManager} );
ok( !$fm->run(), q{testing first failure case} );
my @failed = $fm->get_failed();
is( $#failed, 0, q{one item failed} );
my @succeeded = $fm->get_succeeded();
is ( $#succeeded, 2, q{three items succeeded} );

$fm = EFS::ForkManager->new({
    list   => [qw( a b c d )],
    action => sub {
        my ($arg) = shift;
        die qq{ARGHH!!!\n} if $arg eq q{a};
        return 0 if $arg eq q{b};
        exit 1 if $arg eq q{c};
        return 1;
    },
});
isa_ok( $fm, q{EFS::ForkManager} );
ok( !$fm->run(), q{testing second failure case} );
@failed = sort $fm->get_failed();
is( $#failed, 2, q{three items failed} );
ok( $failed[0] eq q{a} && $failed[1] eq q{b} && $failed[2] eq q{c},
    q{got the correct list of failed elements} );
@succeeded = $fm->get_succeeded();
is( $#succeeded, 0, q{one item succeeded} );
ok( $succeeded[0] eq q{d}, q{got the correct list of succeeded elements} );

$fm = EFS::ForkManager->new({
    list   => [],
    action => sub { 1; },
});
isa_ok( $fm, q{EFS::ForkManager} );
ok( $fm->run(), q{testing noop case (empty list)} );

my $pfm = Parallel::ForkManager->new;

if ( $pfm->can(q{set_retrycount}) ) {

    enable_fork_failure();

    $fm = EFS::ForkManager->new({
        list       => [q{a}],
        action     => sub { 1; },
        retrycount => 0,
    });
    throws_ok {
        $fm->run();
    } qr{Cannot fork: Cannot allocate memory}ms,
        q{simulated fork failure, no retry};

    enable_fork_failure();

    $fm = EFS::ForkManager->new({
        list       => [q{a}],
        action     => sub { 1; },
        retrycount => 1,
        retrysleep => 1,
    });
    ok( $fm->run(), q{simulating fork failure, with retry} );

}

#
# This test hasn't failed in 3 years, and it takes forever to run...
#
# $fm = EFS::ForkManager->new({
#     list   => [ 0 .. 1023 ],
#     action => sub { 1; },
# });
# ok( $fm->run(), q{verifying no file handle leaks} );

$fm = EFS::ForkManager->new({
    list   => [ 0 .. 1 ],
    action => sub { 1; },
});
ok( $fm->get_max_parallel(1) == 8,
    q{verifying parallel of 8 for small processes} );
ok( $fm->get_max_parallel( 1_000_000_001 ) == 4,
    q{verifying parallel of 4 for medium processes} );
ok( $fm->get_max_parallel( 2_000_000_001 ) == 2,
    q{verifying parallel of 2 for large processes} );

exit 0;
