#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 22;
use Test::Exception;

use lib q{test/lib};
use EFS::Test::Config;

use blib;
use EFS;
use EFS::FQDN qw(fqdn);

my $config = EFS::Test::Config->new;

## no critic: RequireExtendedFormatting

my $ldc_efsserver = $config->get_value( q{LocalDevCell}, q{efsserver} );

SKIP: {

    if ( not $config->get_value( q{Security}, q{kerberos_realm} ) ) {
        skip q{Kerberos tests not configured}, 1;
    }

    ## no critic: RequireInterpolationOfMetachars
    my $krb5ccname = qq{/var/tmp/krb5_efstest_$PID};
    $ENV{KRB5CCNAME} = qq{FILE:$krb5ccname};
    system qq{kinit -k -t /var/kerberos/keytabs/efsops.keytab efsops};
    ok( $CHILD_ERROR == 0, q{Acquired efsops credentials for this test} );

}

# Just want to shut up the various messages on stderr...
ok( open( STDERR, q{>&STDOUT} ), q{stderr -> stdout} );

throws_ok {
    EFS::DirLock->new({});
} qr{attribute.*required}ims,
    q{testing parameter validation};

my $directory = q{/var/tmp/efs_dirlock_test};

if ( not -d $directory ) {
    mkdir( $directory, 0755 ) || die qq{Unable to mkdir $directory: $ERRNO\n};
}

my $destroyed = qq{$directory/.destroyed};

if ( not -d $destroyed ) {
    mkdir( $destroyed, 0755 ) ||
        die qq{Unable to mkdir $destroyed: $ERRNO\n};
}

my $first = EFS::DirLock->new({
    directory => $directory,
    command   => q{EFS Test Suite, 1st lock},
});
isa_ok( $first, q{EFS::DirLock} );

throws_ok {
    EFS::DirLock->new({
        directory => $directory,
        command   => q{EFS Test Suite, 2nd lock},
    });
} qr{^Unable to lock $directory}ms,
    q{verifying 2nd lock fails};

throws_ok {
    EFS::DirLock->new({
        directory  => $directory,
        command    => q{EFS Test Suite, 2nd lock},
        retry      => 1,
        retrycount => 1,
        retrysleep => 0,
    });
} qr{^Unable to lock $directory}ms,
    q{verifying 2nd lock fails, with one retry};

ok( $first->destroy(), q{explicit destroy} );

ok( ! -f qq{$directory/.lock}, q{lock removed correctly, explicitly} );

# Test that the lockfile gets removed by the object destructor correctly

{

    my $second = EFS::DirLock->new({
        directory => $directory,
        command   => q{EFS Test Suite, 3rd lock},
    });

    isa_ok( $second, q{EFS::DirLock} );

    ok( -f qq{$directory/.lock}, q{3rd lock file created correctly} );

}

ok( ! -f qq{$directory/.lock}, q{3rd lock removed correctly, implicitly} );

# Test that the lockfile gets removed when moved into the alternate
# directory

{

    my $second = EFS::DirLock->new({
        directory => $directory,
        alternate => $destroyed,
        command   => q{EFS Test Suite, 4th lock},
    });

    isa_ok( $second, q{EFS::DirLock} );

    rename( qq{$directory/.lock}, qq{$destroyed/.lock} ) ||
        die qq{Unable to rename $directory/.lock -> .destroyed/.lock: $ERRNO\n};

}

ok( ! -f qq{$destroyed/.lock}, q{destroyed 4th lock file removed correctly} );

# Test reclaiming a stale lockfile when the efsserver is bogus

system(qq{echo nosuchhost:-1:bogus_command > $directory/.lock});
$CHILD_ERROR and die qq{Unable to create bogus lockfile\n};

{

    my $second = EFS::DirLock->new({
        directory => $directory,
        command   => q{EFS Test Suite, 5th lock},
    });

    isa_ok( $second, q{EFS::DirLock} );

}

ok( ! -f qq{$directory/.lock}, q{5th lock removed correctly, implicitly} );

# Test reclaiming a stale lockfile when the process ain't running locally

my $fqdn = fqdn();

system( qq{echo $fqdn:-1:bogus_command > $directory/.lock} );
$CHILD_ERROR and die qq{Unable to create bogus lockfile\n};

{

    my $second = EFS::DirLock->new({
        directory => $directory,
        command   => q{EFS Test Suite, 6th lock},
    });

    isa_ok( $second, q{EFS::DirLock} );

}

ok( ! -f qq{$directory/.lock}, q{6th lock removed correctly, implicitly} );

# Test reclaiming a stale lockfile when the process ain't running remotely

my $remote = $config->get_value( q{RemoteDevCell}, q{efsserver} );

system( qq{echo $remote:-1:bogus_command > $directory/.lock} );
$CHILD_ERROR and die qq{Unable to create bogus lockfile\n};

{

    # NOTE: You have to use retry here, 'cause the NFS cache can hang
    # on to the .lock after it's been unlinked remotely.

    my $second = EFS::DirLock->new({
        directory => $directory,
        retry     => 1,
        command   => q{EFS Test Suite, 7th lock},
    });

    isa_ok( $second, q{EFS::DirLock} );

}

ok( ! -f qq{$directory/.lock}, q{7th lock removed correctly, implicitly} );

# Test that we can NOT reclaim a lockfile when the process IS running
# remotely.  Note that we're using init's pid (1) which we can be
# fairly certain is still running.

system( qq{echo $remote:1:bogus_command > $directory/.lock} );
$CHILD_ERROR and die qq{Unable to create bogus lockfile\n};

throws_ok {
    EFS::DirLock->new({
        directory => $directory,
        command   => q{EFS Test Suite, 8th lock},
    });
} qr{^Unable to lock $directory}ms,
    q{verifying 8th lock fails};

unlink(qq{$directory/.lock}) ||
    die qq{Unable to unlink $directory/.lock: $ERRNO\n};

# Test the lockname arguemnt

my $lockname = EFS::DirLock->new({
    directory => $directory,
    lockname  => q{.lock-customname},
    command   => q{EFS Test Suite, 9th lock},
});

ok( -f qq{$directory/.lock-customname}, q{9th custom lockname works} );

$lockname->destroy;

ok( ! -f qq{$directory/.lock-customname},
    q{9th custom lockname destroyed correctly} );

exit 0;

END {
    my $krb5ccname = qq{/var/tmp/krb5_efstest_$PID};
    if ( $krb5ccname && -f $krb5ccname ) {
        system q{kdestroy};
        $CHILD_ERROR and warn qq{Unable to destroy kerberos credentials\n};
    }
}
