#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 23;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;

my $mpr_name = q{testmeta testproj testrel};
my $mpr_path = q{testmeta/testproj/testrel};

my @setup = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
    qq{efs create release $mpr_name},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

$test->efs_test_command(
    name    => q{checkpoint release with no installs},
    command => qq{efs checkpoint release $mpr_name -nolock},
    expect  => qq{No installs have been created for release '$mpr_path'},
    not     => 1,
);

$test->efs_test_command(
    name    => q{create a common install},
    command => qq{efs create install $mpr_name common},
    bail    => 1,
);

# XXX: This is a bug!!! We should be locking down the release,
# regardless of whether or not the installs are empty!!

$test->efs_test_command(
    name    => q{checkpoint the empty install},
    command => qq{efs checkpoint release $mpr_name},
    expect  => [
        qq{Checktime not updated for empty install '$mpr_path/common'.},
        qq{Locking directory /efs/dev/$mpr_path/src},
        qq{Locking directory /efs/dev/$mpr_path/build},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

# XXX: This should fail, not succeed!!!

$test->efs_test_command(
    name    => q{create a file in the install},
    command => qq{touch /efs/dev/$mpr_path/install/common/file.txt},
);

$test->efs_test_command(
    name    => q{checkpoint -nolock},
    command => qq{efs checkpoint release -nolock $mpr_name},
    expect  => [
        qq{Performing dependencies audit on install $mpr_path/common},
        qq{Performing forbidden_files audit on install $mpr_path/common},
        qq{Updated initial checktime for install '$mpr_path/common'},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

$test->efs_test_command(
    name    => q{checkpoint -nolock a second time},
    command => qq{efs checkpoint release -nolock $mpr_name},
    expect  => [
        qr{Skipping \s+ dependencies \s+ audit \s+ for \s+ '$mpr_path/common' }msx,
        qr{Skipping \s+ forbidden_files \s+ audit \s+ for \s+ '$mpr_path/common' }msx,
        qq{Unmodified install '$mpr_path/common' skipped.},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

$test->efs_test_command(
    name    => q{change file},
    command => qq{touch /efs/dev/$mpr_path/install/common/file.txt},
);

$test->efs_test_command(
    name    => q{checkpoint modified},
    command => qq{efs checkpoint release $mpr_name},
    expect  => [
        qq{Locking install /efs/dev/$mpr_path/install/common},
        qq{Performing dependencies audit on install $mpr_path/common},
        qq{Performing forbidden_files audit on install $mpr_path/common},
        qr{The \s+ last \s+ checktime \s+ for \s+ this \s+ install \s+ was \s+ .* }msx,
        q{The following paths have been modified since then: },
        qr{\[ \d{4}(-\d\d){2} \s+ \d\d:\d\d:\d\d \] \s+ => \s+ /efs/dev/$mpr_path/install/common}msx,
        qr{\[ \d{4}(-\d\d){2} \s+ \d\d:\d\d:\d\d \] \s+ => \s+ /efs/dev/$mpr_path/install/common/file.txt}msx,
        qq{Updated checktime for modified install '$mpr_path/common'.},
        qq{Locking directory /efs/dev/$mpr_path/src},
        qq{Locking directory /efs/dev/$mpr_path/build},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

$test->efs_test_command(
    name    => q{unlock release},
    command => qq{efs unlock release $mpr_name},
    expect  => [
        qq{Unlocking directory /efs/dev/$mpr_path/src},
        qq{Unlocking directory /efs/dev/$mpr_path/build},
        qq{Unlocking all installs in release '$mpr_path'},
        qq{Unlocking install /efs/dev/$mpr_path/install/common},
        qq{Successfully chowned all installs in $mpr_path},
    ],
);

$test->efs_test_command(
    name    => q{remove file},
    command => qq{rm /efs/dev/$mpr_path/install/common/file.txt},
);

$test->efs_test_command(
    name    => q{checkpoint},
    command => qq{efs checkpoint release $mpr_name -nolock},
    expect  => [
        qq{Removing bogus checktime for empty, unreleased install '$mpr_path/common'.},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

$test->efs_test_command(
    name    => q{again, create a file in the install},
    command => qq{touch /efs/dev/$mpr_path/install/common/file.txt},
);

$test->efs_test_command(
    name    => q{checkpoint (allow lock)},
    command => qq{efs checkpoint release $mpr_name},
    expect  => [
        qq{Locking install /efs/dev/$mpr_path/install/common},
        qq{Performing dependencies audit on install $mpr_path/common},
        qq{Performing forbidden_files audit on install $mpr_path/common},
        qq{Updated initial checktime for install '$mpr_path/common'},
        qq{Locking directory /efs/dev/$mpr_path/src},
        qq{Locking directory /efs/dev/$mpr_path/build},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

$test->efs_test_command(
    name    => q{dist release},
    command => qq{efs dist release $mpr_name},
    filter  => qr{Validating|to \s+ all}mxs,
    expect  => [
        qq{Validating the checkpoint state for release '$mpr_path'.},
        q{Successfully distributed release to all cells},
    ],
);

$test->efs_test_command(
    name    => q{unlock release},
    command => qq{efs unlock release $mpr_name},
);

$test->efs_test_command(
    name    => q{change file},
    command => qq{chmod a+x /efs/dev/$mpr_path/install/common/file.txt},

);

$test->efs_test_command(
    name    => q{checkpoint release},
    command => qq{efs checkpoint release $mpr_name},
    expect  => [
        qq{Locking install /efs/dev/$mpr_path/install/common},
        qq{Performing dependencies audit on install $mpr_path/common},
        qq{Performing forbidden_files audit on install $mpr_path/common},
        qr{The \s+ last \s+ checktime \s+ for \s+ this \s+ install \s+ was}msx,
        q{The following paths have been modified since then:},
        qr{=> \s+ /efs/dev/$mpr_path/install/common}msx,
        qr{=> \s+ /efs/dev/$mpr_path/install/common/file.txt}msx,
        qq{Updated checktime for modified install '$mpr_path/common'.},
        qq{Locking directory /efs/dev/$mpr_path/src},
        qq{Locking directory /efs/dev/$mpr_path/build},
        qq{Successfully completed checkpoint for '$mpr_path'},
    ],
);

my @teardown = (
    qq{efs setstage release $mpr_name rejected},
    qq{efs destroy release $mpr_name},
    qq{efs purge release $mpr_name -now},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;


