#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 39;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;
my $config = $test->get_config;

my $region = $config->get_value( q{Global}, q{region} );

my $ldc_efsserver = $config->get_value( q{LocalDevCell},  q{efsserver} );
my $rdc_efsserver = $config->get_value( q{RemoteDevCell}, q{efsserver} );

my $ldc_name = $config->get_value( q{LocalDevCell},  q{name} );
my $lpc_name = $config->get_value( q{LocalProdCell}, q{name} );
my $rdc_name = $config->get_value( q{RemoteDevCell}, q{name} );

my $ldc_shortname = $config->get_value( q{LocalDevCell},  q{shortname} );
my $lpc_shortname = $config->get_value( q{LocalProdCell}, q{shortname} );
my $rdc_shortname = $config->get_value( q{RemoteDevCell}, q{shortname} );

my @setup = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

$test->efs_test_command(
    name    => q{sanity check display cell (no --hiddencells)},
    command => q{efs display cell},
    filter  => qr{\A Region}xms,
    expect  => [
        qr{Parentcell \s+ Disabled \s* \z}xms,
        qr{Parentcell \s+ Disabled \s* \z}xms,
    ],
);

$test->efs_test_command(
    name    => q{sanity check display cell (with --hiddencells)},
    command => q{efs display cell --hiddencells},
    filter  => qr{\A (Region | $region) \s}xms,
    expect  => [
        qr{\A Region .* Parentcell \s+ Disabled \s+ Hidden \s* \z}xms,
        qr{\A $region .* \s 0 \s* \z}xms,
        qr{\A $region .* \s 0 \s* \z}xms,
        qr{\A Region .* Parentcell \s+ Disabled \s+ Hidden \s* \z}xms,
        qr{\A $region .* \s 0 \s* \z}xms,
    ],
);

## Test efs status efsserver -- should show both

$test->efs_test_command(
    command => q{efs status efsserver},
    expect  => [
        qq{No active EFS commands on $ldc_efsserver (test, $ldc_shortname,$lpc_shortname)},
        qq{No active EFS commands on $rdc_efsserver (test, $rdc_shortname)},
    ],
);

## Test hide cell

$test->efs_test_command(
    run_as  => q{efsops},
    name    => qq{hide $rdc_shortname},
    command => qq{efs hide cell $rdc_shortname},
    bail    => 1,
);

## Test display cell (with hiding) with and without --hiddencells

$test->efs_test_command(
    name    => q{display cell does not show hidden cell},
    command => q{efs display cell --celltype dev},
    filter  => qr{\A $region \s}xms,
    expect  => qr{ $ldc_shortname }xms,
);

$test->efs_test_command(
    name    => q{display cell with --hiddencells shows hidden cell},
    command => q{efs display cell --celltype dev --hiddencells},
    filter  => qr{\A $region \s}xms,
    expect  => [
        qr{ $ldc_shortname .* 0 \s* \z}xms,
        qr{ $rdc_shortname .* 1 \s* \z}xms,
    ],
);

## Test efs status efsserver -- should ignore rdc_efsserver

$test->efs_test_command(
    command => q{efs status efsserver},
    expect  => qq{No active EFS commands on $ldc_efsserver (test, $ldc_shortname,$lpc_shortname)},
);

## Test dist

make_distable_release(q{only-exposed});

$test->efs_test_command(
    name    => q(dist a release without --hiddencells),
    command => q{efs dist release testmeta testproj only-exposed --global},
);

make_distable_release(q{with-hidden});

$test->efs_test_command(
    name    => q(dist a second release with --hiddencells),
    command => q{efs dist release testmeta testproj with-hidden --hiddencells --global},
);

$test->efs_test_command(
    name    => q(Check diststate of first release without --hiddencells),
    command => q{efs diststate release testmeta testproj only-exposed --celltype dev},
    expect  => q{Dev environment is up to date.},
);

$test->efs_test_command(
    name    => q(Check diststate of first release with --hiddencells),
    command => q{efs diststate release testmeta testproj only-exposed --celltype dev --hiddencells},
    filter  => qr{\A (Install | common)}xms,
    expect  => [
        qr{Install \s+ $ldc_shortname \s+ $rdc_shortname}xms,
        qr{common \s+ D \s+ X}xms,
    ],
);

$test->efs_test_command(
    name    => q(Check diststate of second release with --hiddencells),
    command => q{efs diststate release testmeta testproj with-hidden --celltype dev --hiddencells},
    expect  => q{Dev environment is up to date.},
);

## expose cell again

$test->efs_test_command(
    run_as  => q{efsops},
    name    => qq{expose $rdc_shortname},
    command => qq{efs expose cell $rdc_shortname},
    bail    => 1,
);

$test->efs_test_command(
    name    => q{display cell shows cell exposed again},
    command => q{efs display cell --celltype dev},
    filter  => qr{\A $region \s}xms,
    expect  => [ qr{ $ldc_shortname }xms, qr{ $rdc_shortname }xms, ],
);

$test->efs_test_command(
    name    => q(Check diststate of first release without --hiddencells),
    command => q{efs diststate release testmeta testproj only-exposed --celltype dev},
    filter  => qr{\A (Install | common)}xms,
    expect  => [
        qr{Install \s+ $ldc_shortname \s+ $rdc_shortname}xms,
        qr{common \s+ D \s+ X}xms,
    ],
);

$test->efs_test_command(
    name    => q(redisting release to exposed cell),
    command => q{efs dist release testmeta testproj only-exposed --global},
);

$test->efs_test_command(
    name    => q(Check diststate of release to ensure disted to exposed cell),
    command => q{efs diststate release testmeta testproj only-exposed --celltype dev -verbose},
    filter  => qr{\A (Install | common)}xms,
    expect  => [
        qr{Install \s+ $ldc_shortname \s+ $rdc_shortname}xms,
        qr{common \s+ D \s+ D}xms,
    ],
);

# parent/child rules

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs hide cell $ldc_shortname},
    not     => 1,
    expect  => [
        q{The following child cells are not yet hidden:},
        qq{$lpc_name},
        qq{Unable to hide cell $ldc_name},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs hide cell $lpc_shortname},
    expect  => qq{Cell $lpc_name has been hidden},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs hide cell $ldc_shortname},
    expect  => qq{Cell $ldc_name has been hidden},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs expose cell $lpc_shortname},
    not     => 1,
    expect  => [
        qq{Parentcell $ldc_name is hidden},
        qq{Unable to expose $lpc_name},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs expose cell $ldc_shortname},
    expect  => qq{Cell $ldc_name has been exposed},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs expose cell $lpc_shortname},
    expect  => qq{Cell $lpc_name has been exposed},
);

my @teardown = (
    q{efs setstage release testmeta testproj only-exposed rejected},
    q{efs destroy release testmeta testproj only-exposed},
    q{efs purge release testmeta testproj only-exposed -now},
    q{efs setstage release testmeta testproj with-hidden rejected},
    q{efs destroy release testmeta testproj with-hidden},
    q{efs purge release testmeta testproj with-hidden -now},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

# Counts as 4 tests
sub make_distable_release {

    my ($release) = shift;

    $test->efs_test_command(
        command => qq{efs create release testmeta testproj $release},
    );

    $test->efs_test_command(
        command => qq{efs create install testmeta testproj $release common},
    );


    $test->efs_test_command(
        command => qq{touch /efs/dev/testmeta/testproj/$release/install/common/notempty},
    );

    $test->efs_test_command(
        command => qq{efs checkpoint release testmeta testproj $release -nolock},
    );

    return;

}
