#!/usr/bin/env perl

## no critic: RequireExtendedFormatting

use strict;
use warnings;
use English;

use File::Slurp;

use lib q{test/lib};
use EFS::Test;
use EFS::Test::Config;

use Test::More tests => 125;

my $test   = EFS::Test->new;
my $config = EFS::Test::Config->new;

my $upload_server   = $config->get_value( qw( Packaging upload_server ) );
my $upload_username = $config->get_value( qw( Packaging upload_username ) );
my $upload_sshkey   = $config->get_value( qw( Packaging upload_sshkey ) );
my $upload_root     = $config->get_value( qw( Packaging upload_root ) );

my $download_url    = $config->get_value( qw( Packaging download_url ) );
my ($download_dir)  = ( $download_url =~ m{^file://localhost(.*)}msx );

my @upload_projects   = qw( uploada uploadb uploadc );
my @download_projects = qw( downloada downloadb downloadc );

purge_download_dirs(
    stage    => q{setup, phase one},
    projects => [ @upload_projects, @download_projects ],
);

my @setup = (
    q{efs create metaproj testpkgs efsops efsops},
);

my $release       = q{1.0.0-build001};
my $releasealias  = q{1.0.0};
my $redirectalias = q{1.0.0-redirect};
my $aliasalias    = q{1.0};

foreach my $project ( @upload_projects ) {

    push @setup, (
        qq{efs create project testpkgs $project},
        qq{efs create release testpkgs $project $release},
    );

    foreach my $install ( qw( common x86-64.rhel.5 x86-32.rhel.5 ) ) {

        push @setup, (
            qq{efs create install testpkgs $project $release $install},
        );

        my $install_dir = qq{/efs/dev/testpkgs/$project/$release/install/};
        $install_dir .= $install eq q{common} ? $install : qq{.exec/$install};

        push @setup, qq{touch $install_dir/notempty};

    }

    my $src_dir = qq{/efs/dev/testpkgs/$project/$release/src};
    push @setup, qq{touch $src_dir/notempty};

    push @setup, (
        qq{efs create releasealias testpkgs $project $releasealias $release},
    );

    if ( $project eq q{uploadc} ) {
        push @setup, (
            qq{efs create releasealias testpkgs $project $aliasalias $releasealias},
            qq{efs create releasealias testpkgs $project $redirectalias $releasealias},
        );
    }

}

# testpkgs/uploada/1.0.0-build001
#     testpkgs/uploadb/1.0.0
#         testpkgs/uploadb/1.0.0-build001
#             testpkgs/uploadc/1.0.0-redirect
#                 testpkgs/uploadc/1.0.0
#                     testpkgs/uploadc/1.0.0-build001
#     testpkgs/uploadc/1.0
#         testpkgs/uploadc/1.0.0 (repeat)

push @setup, (
    qq{efs define dependency testpkgs uploada $release testpkgs uploadb $releasealias runtime},
    qq{efs define dependency testpkgs uploadb $release testpkgs uploadc $redirectalias runtime},
    qq{efs define dependency testpkgs uploada $release testpkgs uploadc $aliasalias runtime},
);

foreach my $command ( @setup ) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(setup, phase one) $command},
        command => $command,
        bail    => 1,
    );
}

# We need to force the addition of the host key, so that subsequent
# commands have predictable output
$test->efs_test_command(
    run_as  => q{root},
    name    => q{(setup, phase one) Initializing ssh host key},
    command => qq{ssh -o StrictHostKeyChecking=no -i $upload_sshkey -l $upload_username $upload_server true},
    bail    => 1,
);

# End of setup

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $release},
    not     => 1,
    expect  =>[
        qq{Release testpkgs/uploadc/$release is not configured for upload},
        q{All 4 of the following attributes must be defined:},
        q{upload_server},
        q{upload_username},
        q{upload_sshkey},
        q{upload_root},
        q{for either the metaproj, or the project},
    ],
);

my %attrs = (
    upload_server   => $upload_server,
    upload_username => $upload_username,
    upload_sshkey   => $upload_sshkey,
    upload_root     => $upload_root,
    upload_hook     => q{true},
);

foreach my $key ( sort keys %attrs ) {
    $test->efs_test_command(
        run_as  => q{efsops},
        command => qq{efs setattr metaproj testpkgs $key $attrs{$key}},
    );
}

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $release},
    not     => 1,
    expect  =>[
        qq{Validating the checkpoint state for release 'testpkgs/uploadc/$release'.},
        qq{Install 'testpkgs/uploadc/$release/common' has no checktime.},
        qq{Install 'testpkgs/uploadc/$release/x86-32.rhel.5' has no checktime.},
        qq{Install 'testpkgs/uploadc/$release/x86-64.rhel.5' has no checktime.},
        qq{The following installs have invalid checktime states:},
        qq{testpkgs/uploadc/$release/common},
        qq{testpkgs/uploadc/$release/x86-32.rhel.5},
        qq{testpkgs/uploadc/$release/x86-64.rhel.5},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs checkpoint release testpkgs uploadc $release --nolock},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $release},
    not     => 1,
    expect  =>[
        qq{Validating the checkpoint state for release 'testpkgs/uploadc/$release'.},
        qq{Unable to package unlocked release testpkgs/uploadc/$release},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs checkpoint release testpkgs uploadc $release},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $release --packageonly},
    expect  => [
        package_release_output( q{uploadc} ),
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $release --packageonly},
    expect  => [
        qr{Release testpkgs/uploadc/$release was packaged on}ms,
        q{Use --force to re-package},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $releasealias --packageonly --force},
    expect  => [
        package_release_output( q{uploadc} ),
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $releasealias},
    expect  => [
        qr{Release testpkgs/uploadc/$release was packaged on}ms,
        q{Use --force to re-package},
        qq{Uploaded testpkgs/uploadc/$release package to $upload_server},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploadc $releasealias},
    expect  => [
        skip_release_output( q{uploadc} ),
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs checkpoint release testpkgs uploadb $release},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs checkpoint release testpkgs uploada $release},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs uploada $releasealias --recurse},
    expect  => [
        skip_release_output( q{uploadc} ),
        skip_release_output( q{uploadc} ),
        package_release_output( q{uploadb} ),
        qq{Uploaded testpkgs/uploadb/$release package to $upload_server},
        package_release_output ( q{uploada} ),
        qq{Uploaded testpkgs/uploada/$release package to $upload_server},
    ],
);

# End of upload tests

convert_upload_to_download();

# Start of download tests

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloadc $releasealias},
    not     => 1,
    expect  => [
        qq{Metaproj testpkgs is not configured for download},
        q{The following attribute must be defined:},
        q{download_url},
        q{for either the metaproj, or the project},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs setattr metaproj testpkgs download_url $download_url},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs nosuch release},
    not     => 1,
    filter  => qr{No such project}ms,
    expect  => [
        qr{No such project for testpkgs/nosuch/release at $download_url}ms,
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloada nosuch},
    not     => 1,
    filter  => qr{No such release}ms,
    expect  => [
        qr{No such release for testpkgs/downloada/nosuch at $download_url}ms,
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloadc $releasealias},
    filter  => qr{Downloaded release}ms,
    expect  => [
        qr{Downloaded release testpkgs/downloadc/$release},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloadc $releasealias},
    expect  => [
        qr{Release testpkgs/downloadc/$release was downloaded on}ms,
        q{Use --force to re-download},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloadc $releasealias --force},
    not     => 1,
    expect  => [
        qq{Release testpkgs/downloadc/$release is locked},
        q{Unable to download a locked release},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs unlock release testpkgs downloadc $release},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloadc $releasealias --force},
    filter  => qr{Downloaded release}ms,
    expect  => [
        qr{Downloaded release testpkgs/downloadc/$release},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs download release testpkgs downloada $releasealias},
    filter  => qr{Downloaded release}ms,
    expect  => [
        qr{Downloaded release testpkgs/downloadb/$release},
        qr{Downloaded release testpkgs/downloada/$release},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs showtree dependency testpkgs downloada $releasealias},
    expect  => [
        qq{Stage Depth Diststate Runtime Dependencies},
        qr{=+}ms,
        qq{dev 1 1d testpkgs/downloada/$releasealias},
        qq{dev 2 1d testpkgs/downloada/$release},
        qq{dev 3 1d testpkgs/downloadb/$releasealias},
        qq{dev 4 1d testpkgs/downloadb/$release},
        qq{dev 5 1d testpkgs/downloadc/$redirectalias},
        qq{dev 6 1d testpkgs/downloadc/$releasealias},
        qq{dev 7 1d testpkgs/downloadc/$release},
        qq{dev 3 1d testpkgs/downloadc/$aliasalias},
        qq{dev 4 1d testpkgs/downloadc/$releasealias (repeat)},
    ],
);

# Start of teardown

my @teardown = ();

foreach my $type ( qw( upload download ) ) {
    push @teardown, (
        qq{efs undef dependency testpkgs ${type}a $release testpkgs ${type}c $aliasalias runtime},
        qq{efs undef dependency testpkgs ${type}b $release testpkgs ${type}c $redirectalias runtime},
        qq{efs undef dependency testpkgs ${type}a $release testpkgs ${type}b $releasealias runtime},
        qq{efs destroy releasealias testpkgs ${type}c $aliasalias},
        qq{efs destroy releasealias testpkgs ${type}c $redirectalias},
    );
}

foreach my $project ( @upload_projects, @download_projects ) {
    push @teardown, (
        qq{efs destroy releasealias testpkgs $project $releasealias},
        qq{efs setstage release testpkgs $project $release rejected},
        qq{efs destroy release testpkgs $project $release},
        qq{efs purge release testpkgs $project $release --now},
        qq{efs destroy project testpkgs $project},
    );
}

push @teardown, q{efs destroy metaproj testpkgs};

foreach my $command ( @teardown ) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $command},
        command => $command,
    );
}

purge_download_dirs(
    stage    => q{teardown},
    projects => [ @upload_projects, @download_projects ],
);

exit 0;

sub purge_download_dirs {

    my %args = @_;

    foreach my $project ( @{ $args{projects} } ) {
        my $dir = qq{$download_dir/testpkgs/$project};
        my $command = qq{su $upload_username -c 'rm -rf $dir'};
        $test->efs_test_command(
            run_as  => q{root},
            name    => qq{($args{stage}) $command},
            command => $command,
            bail    => 1,
        );
    }

    return 1;

}

sub package_release_output {

    my $project = shift;

    my $directory = qq{/efs/dev/testpkgs/$project/$release/build/upload};

    my @output = (
        qq{Validating the checkpoint state for release 'testpkgs/$project/$release'.},
        qr{(Removed|No such) directory $directory}ms,
        qq{Created directory $directory},
        qq{Created directory $directory/package},
        qq{Created directory $directory/package/install},
        qq{Creating $directory/package/install/common.tar.bz2},
        qq{Creating $directory/package/install/x86-32.rhel.5.tar.bz2},
        qq{Creating $directory/package/install/x86-64.rhel.5.tar.bz2},
        qq{Created directory $directory/aliases},
    );

    if ( $project eq q{uploadc} ) {
        push @output, (
            qq{Created symlink $directory/aliases/$aliasalias -> $releasealias},
        );
    }

    push @output, (
        qq{Created symlink /efs/dev/testpkgs/$project/$release/build/upload/aliases/$releasealias -> $release},
    );

    if ( $project eq q{uploadc} ) {
        push @output, (
            qq{Created symlink $directory/aliases/$redirectalias -> $releasealias},
        );
    }

    push @output, (
        qq{Packaged testpkgs/$project/$release},
    );

    return @output;

}

sub skip_release_output {

    my $project = shift;

    return (
        qr{Release testpkgs/$project/$release was packaged on}ms,
        q{Use --force to re-package},
        qr{Release testpkgs/$project/$release was uploaded on}ms,
        q{Use --force to re-upload},
    );

}

sub convert_upload_to_download {

    my @setup = ();

    foreach my $upload ( @upload_projects ) {

        my $download = $upload;
        $download =~ s{upload}{download}gmsx;

        my $rsync = qq{rsync -az $download_dir/testpkgs/$upload/ $download_dir/testpkgs/$download};
        my $command = qq{su $upload_username -c '$rsync'};

        $test->efs_test_command(
            run_as  => q{root},
            name    => qq{(setup, phase two) $command},
            command => $command,
            bail    => 1,
        );

        my $file = qq{$download_dir/testpkgs/$download/$release/metadata.conf};
        my $temp = qq{/var/tmp/testpkgs-$download-$release-metadata.conf};

        if ( -f $temp ) {
            unlink $temp || die qq{Unable to unlink $temp: $ERRNO\n};
        }

        my $metadata = read_file( $file );
        $metadata =~ s{upload}{download}gmsx;
        write_file( $temp, $metadata );

        $test->efs_test_command(
            run_as  => q{root},
            name    => qq{(setup, phase two) edit $file},
            command => qq{su $upload_username -c 'cp $temp $file'},
        );

        unlink $temp || die qq{Unable to unlink $temp: $ERRNO\n};

    }

}
