#!/usr/bin/env perl

use 5.010;
use warnings;
use strict;
use English;

=head1 DESCRIPTION

Makes sure that all necessary modules (loaded via C<use> or C<require>) are listed in the
F<Makefile.PL> as dependencies.

=cut

use Test::More;
use File::Find;
use Readonly;
use IO::File;

eval {
    use Module::CoreList;
};
if ($EVAL_ERROR) {
    plan skip_all => qq{Trouble loading Module::CoreList:\n$EVAL_ERROR}
}
elsif ( ! -e q{inc/.author} ) {
    plan skip_all => q{no inc/.author, and it is on purpose ;)}
}

plan q{no_plan};

Readonly my @IGNORED_MODULES => qw/ DBIx::Class::Schema /;

my %used;
find( \&wanted, qw/ lib bin t test xt / );

sub wanted {
    return unless -f $_;
    return if $File::Find::dir =~ m{/\.(git|svn)($|/)}msx;
    return if $File::Find::name =~ /~$/msx;
    return if $File::Find::name =~ /\.(pod|html)$/msx;

    ##  read in the file from disk
    my $filename = $_;
    local $RS = undef;
    my $fh = IO::File->new($filename) || do {
        warn qq{Unable to open $filename: $ERRNO\n};
        ok( 0, qq{Opened $filename} );
        return;
    };
    my $data = <$fh>;
    $fh->close;

    ##  strip pod, in a really idiotic way. Good enough though
    $data =~ s/^=head.+?(^=cut|\Z)//gmsx;

    ##  look for 'use', 'use parent', and 'use base' statements
    while ( $data =~ /^\s*use\s+([[:alpha:]][\w:]+)/gmsx ) {
        $used{$1}{$File::Find::name}++;
    }

    while ( $data =~ m/^\s*use\s+(?:parent|base)\s+qw.([[:alpha:]][\w\s:]+)/gmsx ) {
        foreach ( split /\s+/msx, $1 ) {
            $used{$_}{$File::Find::name}++;
        }
    }

    return;
}

##  extract required and recommended modules and perl version from Makefile.PL
my $required_perl_version = 0;
my %required;
{
    ## no critic: RequireDotMatchAnything
    ## no critic: RequireExtendedFormatting

    local $RS = undef;
    my $fh = IO::File->new(q{Makefile.PL}) ||
        die qq{Unable to open Makefile.PL: $ERRNO\n};
    my $data = <$fh>;
    $fh->close;

    ##  extract required perl version
    while ( $data =~ m/^\s*?perl_version\s+['"]([0-9._]*)['"].*?$/gm ) {
        if (defined $1 and length $1) {
            $required_perl_version = $1;
            last;
        }
    }

    ##  extract required and recommended modules
    while ( $data =~ /^\s*?(?:requires|recommends)?.*?([\w:]+)'(?:\s*=>\s*['"]?([\d\.]+)['"]?)?.*?(?:#(.*))?$/gm ) {
        $required{$1} = $2;
        if ( defined $3 and length $3 ) {
            foreach ( split /\s+/msx, $3 ) {
                $required{$_} = undef;
            }
        }
    }
}

##  verify all required|recommended modules are listed in Makefile.PL
##  ignore application/local modules
for ( sort keys %used ) {
    my $first_in = Module::CoreList->first_release($_);
    next if defined $first_in and $first_in <= $required_perl_version;
    next if /^(EFS|inc|t)(::|$)/msx;

    ##  ignore modules included in a distro listed in Makefile.PL
    next if $_ ~~ @IGNORED_MODULES;

    #warn $_;
    ok( exists $required{$_}, qq{$_ in Makefile.PL} )
        or diag( q{used in }, join ', ', sort keys %{ $used{$_} } );
    delete $used{$_};
    delete $required{$_};
}

##  don't include known core modules in Makefile.PL
for ( sort keys %required ) {
    my $first_in = Module::CoreList->first_release( $_, $required{$_} );
    fail(qq{Required module $_ (v. $required{$_}) is in core since $first_in})
    if defined $first_in and $first_in <= $required_perl_version;
}

